@@ -21,21 +21,21 @@ namespace Mono.Debugger.Backend
 	{
 		public static TimeSpan WaitTimeout = TimeSpan.FromMilliseconds (5000);
 
-		internal ThreadManager (DebuggerServant backend)
+		internal ThreadManager (Debugger debugger)
 		{
-			this.backend = backend;
+			this.debugger = debugger;
 
 			thread_hash = Hashtable.Synchronized (new Hashtable ());
 			engine_hash = Hashtable.Synchronized (new Hashtable ());
 			processes = ArrayList.Synchronized (new ArrayList ());
 
 			pending_events = Hashtable.Synchronized (new Hashtable ());
+
+			last_pending_sigstop = DateTime.Now;
+			pending_sigstops = new Dictionary<int,DateTime> ();
 			
 			address_domain = AddressDomain.Global;
 
-			command_mutex = new DebuggerMutex ("command_mutex");
-			command_mutex.DebugFlags = DebugFlags.Wait;
-
 			wait_event = new ST.AutoResetEvent (false);
 			engine_event = new ST.ManualResetEvent (true);
 			ready_event = new ST.ManualResetEvent (false);
@@ -56,7 +56,7 @@ namespace Mono.Debugger.Backend
 			ready_event.WaitOne ();
 		}
 
-		DebuggerServant backend;
+		Debugger debugger;
 		DebuggerEventQueue event_queue;
 		ST.Thread inferior_thread;
 		ST.Thread wait_thread;
@@ -70,7 +70,9 @@ namespace Mono.Debugger.Backend
 
 		AddressDomain address_domain;
 
-		DebuggerMutex command_mutex;
+		DateTime last_pending_sigstop;
+		Dictionary<int,DateTime> pending_sigstops;
+
 		bool abort_requested;
 		bool waiting;
 
@@ -80,9 +82,6 @@ namespace Mono.Debugger.Backend
 		[DllImport("monodebuggerserver")]
 		static extern int mono_debugger_server_global_wait (out int status);
 
-		[DllImport("monodebuggerserver")]
-		static extern int mono_debugger_server_get_pending_sigint ();
-
 		[DllImport("monodebuggerserver")]
 		static extern Inferior.ChildEventType mono_debugger_server_dispatch_simple (int status, out int arg);
 
@@ -94,6 +93,8 @@ namespace Mono.Debugger.Backend
 			while (!abort_requested) {
 				engine_thread_main ();
 			}
+
+			Report.Debug (DebugFlags.Threads, "Engine thread exiting.");
 		}
 
 		// <remarks>
@@ -104,19 +105,14 @@ namespace Mono.Debugger.Backend
 		SingleSteppingEngine current_event = null;
 		int current_event_status = 0;
 
-		public ProcessServant StartApplication (ProcessStart start)
-		{
-			ProcessServant process = CreateProcess (start);
-			process.WaitForApplication ();
-			return process;
-		}
-
-		public ProcessServant OpenCoreFile (ProcessStart start, out Thread[] threads)
+#if DISABLED
+		public Process OpenCoreFile (ProcessStart start, out Thread[] threads)
 		{
 			CoreFile core = CoreFile.OpenCoreFile (this, start);
 			threads = core.GetThreads ();
 			return core;
 		}
+#endif
 
 		internal void AddEngine (SingleSteppingEngine engine)
 		{
@@ -124,18 +120,7 @@ namespace Mono.Debugger.Backend
 			engine_hash.Add (engine.ID, engine);
 		}
 
-		internal void ProcessExecd (SingleSteppingEngine engine)
-		{
-			SingleSteppingEngine old_engine = (SingleSteppingEngine) thread_hash [engine.PID];
-			if (old_engine != null) {
-				thread_hash [engine.PID] = engine;
-				engine_hash.Remove (old_engine.ID);
-			} else
-				thread_hash.Add (engine.PID, engine);
-			engine_hash.Add (engine.ID, engine);
-		}
-
-		internal void RemoveProcess (ProcessServant process)
+		internal void RemoveProcess (Process process)
 		{
 			processes.Remove (process);
 		}
@@ -162,10 +147,11 @@ namespace Mono.Debugger.Backend
 				return true;
 			}
 
-			inferior.Process.Initialize (engine, inferior, false);
-
 			if (cevent.Type == Inferior.ChildEventType.CHILD_CREATED_THREAD) {
-				inferior.Process.ThreadCreated (inferior, (int) cevent.Argument, false);
+				int pid = (int) cevent.Argument;
+				inferior.Process.ThreadCreated (inferior, pid, false, true);
+				if (pending_sigstops.ContainsKey (pid))
+					pending_sigstops.Remove (pid);
 				resume_target = true;
 				return true;
 			}
@@ -177,7 +163,9 @@ namespace Mono.Debugger.Backend
 			}
 
 			if (cevent.Type == Inferior.ChildEventType.CHILD_EXECD) {
-				inferior.Process.ChildExecd (inferior);
+				thread_hash.Remove (engine.PID);
+				engine_hash.Remove (engine.ID);
+				inferior.Process.ChildExecd (engine, inferior);
 				resume_target = false;
 				return true;
 			}
@@ -188,6 +176,9 @@ namespace Mono.Debugger.Backend
 						Inferior.ChildEventType.CHILD_STOPPED, 0, 0, 0);
 					resume_target = true;
 					return true;
+				} else if (inferior.Has_SIGWINCH && (cevent.Argument == inferior.SIGWINCH)) {
+					resume_target = true;
+					return true;
 				} else if (inferior.HasSignals && (cevent.Argument == inferior.Kernel_SIGRTMIN+1)) {
 					// __SIGRTMIN and __SIGRTMIN+1 are used internally by the threading library
 					resume_target = true;
@@ -195,6 +186,11 @@ namespace Mono.Debugger.Backend
 				}
 			}
 
+			if (inferior.Process.OperatingSystem.CheckForPendingMonoInit (inferior)) {
+				resume_target = true;
+				return true;
+			}
+
 			bool retval = false;
 			resume_target = false;
 			if (inferior.Process.MonoManager != null)
@@ -202,7 +198,7 @@ namespace Mono.Debugger.Backend
 					engine, inferior, ref cevent, out resume_target);
 
 			if ((cevent.Type == Inferior.ChildEventType.CHILD_EXITED) ||
-			     (cevent.Type == Inferior.ChildEventType.CHILD_SIGNALED)) {
+			    (cevent.Type == Inferior.ChildEventType.CHILD_SIGNALED)) {
 				thread_hash.Remove (engine.PID);
 				engine_hash.Remove (engine.ID);
 				engine.OnThreadExited (cevent);
@@ -213,8 +209,17 @@ namespace Mono.Debugger.Backend
 			return retval;
 		}
 
-		public DebuggerServant Debugger {
-			get { return backend; }
+		internal bool HasPendingSigstopForNewThread (int pid)
+		{
+			if (!pending_sigstops.ContainsKey (pid))
+				return false;
+
+			pending_sigstops.Remove (pid);
+			return true;
+		}
+
+		public Debugger Debugger {
+			get { return debugger; }
 		}
 
 		public AddressDomain AddressDomain {
@@ -249,7 +254,7 @@ namespace Mono.Debugger.Backend
 				return command.Result;
 		}
 
-		internal ProcessServant CreateProcess (ProcessStart start)
+		public Process StartApplication (ProcessStart start, out CommandResult result)
 		{
 			Command command = new Command (CommandType.CreateProcess, start);
 
@@ -268,8 +273,11 @@ namespace Mono.Debugger.Backend
 
 			if (command.Result is Exception)
 				throw (Exception) command.Result;
-			else
-				return (ProcessServant) command.Result;
+			else {
+				var pair = (KeyValuePair<CommandResult,Process>) command.Result;
+				result = pair.Key;
+				return pair.Value;
+			}
 		}
 
 		internal void AddPendingEvent (SingleSteppingEngine engine, Inferior.ChildEvent cevent)
@@ -347,8 +355,9 @@ namespace Mono.Debugger.Backend
 			// until we finished the command and sent the result.
 			if (command.Type == CommandType.TargetAccess) {
 				try {
-					command.Result = command.Engine.Invoke (
-						(TargetAccessDelegate) command.Data1, command.Data2);
+					if(command.Engine.Inferior != null)
+						command.Result = command.Engine.Invoke (
+							(TargetAccessDelegate) command.Data1, command.Data2);
 				} catch (ST.ThreadAbortException) {
 					return;
 				} catch (Exception ex) {
@@ -361,19 +370,14 @@ namespace Mono.Debugger.Backend
 			} else if (command.Type == CommandType.CreateProcess) {
 				try {
 					ProcessStart start = (ProcessStart) command.Data1;
-					ProcessServant process = new ProcessServant (this, start);
-
-					CommandResult result;
-					SingleSteppingEngine sse = new SingleSteppingEngine (
-						this, process, start, out result);
-
-					thread_hash.Add (sse.PID, sse);
-					engine_hash.Add (sse.ID, sse);
+					Process process = new Process (this, start);
 					processes.Add (process);
 
+					CommandResult result = process.StartApplication ();
+
 					RequestWait ();
 
-					command.Result = process;
+					command.Result = new KeyValuePair<CommandResult,Process> (result, process);
 				} catch (ST.ThreadAbortException) {
 					return;
 				} catch (Exception ex) {
@@ -467,15 +471,29 @@ namespace Mono.Debugger.Backend
 				//
 
 				do {
+					Report.Debug (DebugFlags.Wait,
+						      "Wait thread reaping children");
 					pid = mono_debugger_server_global_wait (out status);
 					Report.Debug (DebugFlags.Wait,
 						      "Wait thread received event: {0} {1:x}",
 						      pid, status);
 				} while (pid > 0);
 
+				Report.Debug (DebugFlags.Wait,
+					      "Wait thread done");
+
 				return false;
 			}
 
+			if (DateTime.Now - last_pending_sigstop > new TimeSpan (0, 2, 30)) {
+				foreach (int pending in pending_sigstops.Keys) {
+					Report.Error ("Got SIGSTOP from unknown PID {0}!", pending);
+				}
+
+				pending_sigstops.Clear ();
+				last_pending_sigstop = DateTime.Now;
+			}
+
 			Report.Debug (DebugFlags.Wait, "Wait thread waiting");
 
 			//
@@ -496,10 +514,64 @@ namespace Mono.Debugger.Backend
 			if (abort_requested || (pid <= 0))
 				return true;
 
-			SingleSteppingEngine event_engine = (SingleSteppingEngine) thread_hash [pid];
-			if (event_engine == null) {
+			if(!Inferior.HasThreadEvents)
+			{
 				int arg;
 				Inferior.ChildEventType etype = mono_debugger_server_dispatch_simple (status, out arg);
+				SingleSteppingEngine engine = (SingleSteppingEngine) thread_hash [pid];
+				if(etype == Inferior.ChildEventType.CHILD_EXITED) {
+					if(engine != null) {
+						SingleSteppingEngine[] sses = new SingleSteppingEngine [thread_hash.Count];
+						thread_hash.Values.CopyTo (sses, 0);
+						foreach(SingleSteppingEngine sse in sses)
+							sse.ProcessEvent (status);
+						Dispose();
+						waiting = false;
+						return true;
+					}
+					else
+						goto again;
+				}
+
+				if (engine == null) {
+					SingleSteppingEngine[] sses = new SingleSteppingEngine [thread_hash.Count];
+					thread_hash.Values.CopyTo (sses, 0);				
+					Inferior inferior = sses[0].Inferior;		
+					inferior.Process.ThreadCreated (inferior, pid, false, true);
+					goto again;
+				}
+
+				ArrayList check_threads = new ArrayList();
+				bool got_threads = true;
+				foreach(Process process in processes)
+					got_threads = got_threads && process.CheckForThreads(check_threads);
+
+				if(got_threads) {
+					int[] lwps = new int [thread_hash.Count];
+					thread_hash.Keys.CopyTo (lwps, 0);
+					foreach(int lwp in lwps) {
+						if(!check_threads.Contains(lwp)) {
+							SingleSteppingEngine old_engine = (SingleSteppingEngine) thread_hash [lwp];					
+							thread_hash.Remove (old_engine.PID);
+							engine_hash.Remove (old_engine.ID);
+							old_engine.Process.OnThreadExitedEvent (old_engine);
+							old_engine.Dispose ();
+						}
+					}
+				}
+			}
+
+			SingleSteppingEngine event_engine = (SingleSteppingEngine) thread_hash [pid];
+			if (event_engine == null && Inferior.HasThreadEvents) {
+				int arg;
+				Inferior.ChildEventType etype = mono_debugger_server_dispatch_simple (status, out arg);
+
+				/*
+				 * Ignore exit events from unknown children.
+				 */
+
+				if ((etype == Inferior.ChildEventType.CHILD_EXITED) && (arg == 0))
+					goto again;
 
 				/*
 				 * There is a race condition in the Linux kernel which shows up on >= 2.6.27:
@@ -522,6 +594,9 @@ namespace Mono.Debugger.Backend
 					return true;
 				}
 
+				if (!pending_sigstops.ContainsKey (pid))
+					pending_sigstops.Add (pid, DateTime.Now);
+
 				Report.Debug (DebugFlags.Wait, "Ignoring SIGSTOP from unknown pid {0}.", pid);
 				goto again;
 			}
@@ -588,7 +663,10 @@ namespace Mono.Debugger.Backend
 #if FIXME
 				RequestWait ();
 #endif
-				event_queue.Signal();
+				if (event_queue.TryLock ()) {
+					event_queue.Signal();
+					event_queue.Unlock ();
+				}
 				disposed = true;
 			}
 
@@ -637,7 +715,7 @@ namespace Mono.Debugger.Backend
 				wait_thread.Abort ();
 				wait_thread.Join ();
 
-				ProcessServant[] procs = new ProcessServant [processes.Count];
+				Process[] procs = new Process [processes.Count];
 				processes.CopyTo (procs, 0);
 
 				for (int i = 0; i < procs.Length; i++)
