@@ -17,7 +17,7 @@ namespace Mono.Debugger.Backend
 {
 	internal abstract class ThreadServant : DebuggerMarshalByRefObject
 	{
-		protected ThreadServant (ThreadManager manager, ProcessServant process)
+		protected ThreadServant (ThreadManager manager, Process process)
 		{
 			this.manager = manager;
 			this.process = process;
@@ -27,19 +27,17 @@ namespace Mono.Debugger.Backend
 			tgroup = process.Session.CreateThreadGroup ("@" + ID);
 			tgroup.AddThread (ID);
 
-			thread = process.Debugger.Client.CreateThread (this, ID);
+			thread = process.Debugger.CreateThread (this, ID);
 		}
 
 		protected readonly int id;
 		protected readonly Thread thread;
-		protected readonly ProcessServant process;
+		protected readonly Process process;
 		protected readonly ThreadManager manager;
 		protected readonly ThreadGroup tgroup;
 
-		bool is_daemon;
-
 		protected internal Language NativeLanguage {
-			get { return process.BfdContainer.NativeLanguage; }
+			get { return process.NativeLanguage; }
 		}
 
 		public override string ToString ()
@@ -80,28 +78,18 @@ namespace Mono.Debugger.Backend
 			get;
 		}
 
-		internal abstract ProcessServant ProcessServant {
-			get;
-		}
-
-		internal ProcessServant Process {
+		internal Process Process {
 			get { return process; }
 		}
 
 		public bool IsDaemon {
-			get { return is_daemon; }
+			get { return (thread.ThreadFlags & (Thread.Flags.Daemon | Thread.Flags.Immutable)) != 0; }
 		}
 
 		public ThreadGroup ThreadGroup {
 			get { return tgroup; }
 		}
 
-		internal void SetDaemon ()
-		{
-			is_daemon = true;
-			thread.SetThreadFlags (thread.ThreadFlags | Thread.Flags.Daemon | Thread.Flags.Immutable);
-		}
-
 		public abstract TargetEventArgs LastTargetEvent {
 			get;
 		}
@@ -144,39 +132,9 @@ namespace Mono.Debugger.Backend
 
 		public abstract Backtrace GetBacktrace (Backtrace.Mode mode, int max_frames);
 
-		// <summary>
-		//   Step one machine instruction, but don't step into trampolines.
-		// </summary>
-		public abstract void StepInstruction (CommandResult result);
+		public abstract CommandResult Step (ThreadingModel model, StepMode mode, StepFrame frame);
 
-		// <summary>
-		//   Step one machine instruction, always step into method calls.
-		// </summary>
-		public abstract void StepNativeInstruction (CommandResult result);
-
-		// <summary>
-		//   Step one machine instruction, but step over method calls.
-		// </summary>
-		public abstract void NextInstruction (CommandResult result);
-
-		// <summary>
-		//   Step one source line.
-		// </summary>
-		public abstract void StepLine (CommandResult result);
-
-		// <summary>
-		//   Step one source line, but step over method calls.
-		// </summary>
-		public abstract void NextLine (CommandResult result);
-
-		// <summary>
-		//   Continue until leaving the current method.
-		// </summary>
-		public abstract void Finish (bool native, CommandResult result);
-
-		public abstract void Continue (TargetAddress until, CommandResult result);
-
-		public abstract void Background (TargetAddress until, CommandResult result);
+		internal abstract ThreadCommandResult Old_Step (StepMode mode, StepFrame frame);
 
 		public abstract void Kill ();
 
@@ -186,6 +144,10 @@ namespace Mono.Debugger.Backend
 
 		public abstract void Stop ();
 
+		internal abstract void SuspendUserThread ();
+
+		internal abstract void ResumeUserThread (CommandResult result);
+
 		internal abstract object Invoke (TargetAccessDelegate func, object data);
 
 		// <summary>
@@ -203,37 +165,23 @@ namespace Mono.Debugger.Backend
 		// </summary>
 		internal abstract void RemoveBreakpoint (BreakpointHandle handle);
 
-		// <summary>
-		//   Add an event handler.
-		//
-		//   Returns a number which may be passed to RemoveEventHandler() to remove
-		//   the event handler.
-		// </summary>
-		public abstract int AddEventHandler (Event handle);
-
-		// <summary>
-		//   Remove event handler @index.  @index is the event handler number which has
-		//   been returned by AddEventHandler().
-		// </summary>
-		public abstract void RemoveEventHandler (int index);
-
 		internal abstract void AcquireThreadLock ();
 
 		internal abstract void ReleaseThreadLock ();
 
-		internal abstract void ReleaseThreadLockDone ();
-
 		public abstract string PrintObject (Style style, TargetObject obj, DisplayFormat format);
 
 		public abstract string PrintType (Style style, TargetType type);
 
-		internal abstract Registers GetCallbackFrame (TargetAddress stack_pointer,
-							      bool exact_match);
+		internal abstract Inferior.CallbackFrame GetCallbackFrame (TargetAddress stack_pointer,
+									   bool exact_match);
+
+		internal abstract TargetFunctionType GetRuntimeInvokedFunction (long ID);
 
 		public abstract void RuntimeInvoke (TargetFunctionType function,
 						    TargetStructObject object_argument,
 						    TargetObject[] param_objects,
-						    bool is_virtual, bool debug,
+						    RuntimeInvokeFlags flags,
 						    RuntimeInvokeResult result);
 
 		public abstract CommandResult CallMethod (TargetAddress method, long arg1,
@@ -245,9 +193,9 @@ namespace Mono.Debugger.Backend
 		public abstract CommandResult CallMethod (TargetAddress method, TargetAddress method_argument,
 							  TargetObject object_argument);
 
-		public abstract CommandResult Return (bool run_finally);
+		public abstract CommandResult Return (ReturnMode mode);
 
-		public abstract CommandResult AbortInvocation ();
+		internal abstract void AbortInvocation (long ID);
 
 		public string PrintRegisters (StackFrame frame)
 		{
@@ -270,6 +218,10 @@ namespace Mono.Debugger.Backend
 			get;
 		}
 
+		public abstract ST.WaitHandle WaitHandle {
+			get;
+		}
+
 		public abstract TargetMemoryInfo TargetMemoryInfo {
 			get;
 		}
