@@ -9,18 +9,19 @@ namespace Mono.Debugger.Backend
 	// </summary>
 	internal abstract class Architecture : DebuggerMarshalByRefObject, IDisposable
 	{
-		protected readonly ProcessServant process;
+		protected readonly Process process;
 		protected readonly TargetInfo TargetInfo;
 
-		BfdDisassembler disassembler;
+		Disassembler disassembler;
 		X86_Opcodes opcodes;
 
-		protected Architecture (ProcessServant process, TargetInfo info)
+		protected Architecture (Process process, TargetInfo info)
 		{
 			this.process = process;
 			this.TargetInfo = info;
 
-			disassembler = new BfdDisassembler (process, info.TargetAddressSize == 8);
+			if (!Inferior.IsRunningOnWindows)
+				disassembler = new BfdDisassembler (process, info.TargetAddressSize == 8);
 			if (info.TargetAddressSize == 8)
 				opcodes = new Opcodes_X86_64 (process);
 			else
@@ -124,10 +125,10 @@ namespace Mono.Debugger.Backend
 		internal abstract StackFrame TrySpecialUnwind (StackFrame last_frame,
 							       TargetMemoryAccess memory);
 
-		internal abstract StackFrame CreateFrame (Thread thread, TargetMemoryAccess target,
-							  Registers regs);
+		internal abstract StackFrame CreateFrame (Thread thread, FrameType type,
+							  TargetMemoryAccess target, Registers regs);
 
-		internal StackFrame CreateFrame (Thread thread, TargetMemoryAccess target,
+		internal StackFrame CreateFrame (Thread thread, FrameType type, TargetMemoryAccess target,
 						 TargetAddress address, TargetAddress stack,
 						 TargetAddress frame_pointer, Registers regs)
 		{
@@ -137,11 +138,11 @@ namespace Mono.Debugger.Backend
 			Method method = process.SymbolTableManager.Lookup (address);
 			if (method != null)
 				return new StackFrame (
-					thread, address, stack, frame_pointer, regs, method);
+					thread, type, address, stack, frame_pointer, regs, method);
 
 			Symbol name = process.SymbolTableManager.SimpleLookup (address, false);
 			return new StackFrame (
-				thread, address, stack, frame_pointer, regs,
+				thread, type, address, stack, frame_pointer, regs,
 				process.NativeLanguage, name);
 		}
 
