@@ -45,17 +45,24 @@ namespace Mono.Debugger
 		public readonly int File;
 		public readonly int Line;
 		public readonly bool IsHidden;
+		public readonly SourceRange? SourceRange;
 
 		public LineEntry (TargetAddress address, int file, int line)
 			: this (address, file, line, false)
 		{ }
 
 		public LineEntry (TargetAddress address, int file, int line, bool is_hidden)
+			: this (address, file, line, is_hidden, null)
+		{ }
+
+		public LineEntry (TargetAddress address, int file, int line, bool is_hidden,
+				  SourceRange? source_range)
 		{
 			this.Address = address;;
 			this.File = file;
 			this.Line = line;
 			this.IsHidden = is_hidden;
+			this.SourceRange = source_range;
 		}
 
 		public int CompareTo (object obj)
@@ -166,6 +173,9 @@ namespace Mono.Debugger
 			get {
 				return name;
 			}
+			protected set {
+				name = value;
+			}
 		}
 
 		public string ImageFile {
@@ -292,7 +302,7 @@ namespace Mono.Debugger
                         return true;
                 }
 
-		public abstract TargetStructType GetDeclaringType (Thread target);
+		public abstract TargetClassType GetDeclaringType (Thread target);
 
 		public abstract bool HasThis {
 			get;
