@@ -10,7 +10,7 @@ using System.Collections.Specialized;
 using System.Runtime.InteropServices;
 using System.Runtime.Remoting.Messaging;
 
-using Mono.Debugger.Backends;
+using Mono.Debugger.Backend;
 using Mono.Debugger.Languages;
 
 namespace Mono.Debugger
@@ -18,7 +18,7 @@ namespace Mono.Debugger
 	[Serializable]
 	internal delegate object TargetAccessDelegate (Thread target, object user_data);
 
-	public class Thread : TargetAccess
+	public sealed class Thread : DebuggerMarshalByRefObject, IOperationHost
 	{
 		[Flags]
 		public enum Flags {
@@ -26,28 +26,26 @@ namespace Mono.Debugger
 			Daemon		= 0x0001,
 			Immutable	= 0x0002,
 			Background	= 0x0004,
-			AutoRun		= 0x0008
+			AutoRun		= 0x0008,
+			StopOnExit	= 0x0010
 		}
 
 		internal Thread (ThreadServant servant, int id)
 		{
 			this.id = id;
 			this.servant = servant;
-			this.is_running = true;
-			this.operation_completed_event = new ST.ManualResetEvent (false);
+			this.flags = Flags.AutoRun;
 		}
 
 		int id;
 		Flags flags;
-		bool is_running;
-		ST.ManualResetEvent operation_completed_event;
 		ThreadServant servant;
 
 		public ST.WaitHandle WaitHandle {
-			get { return operation_completed_event; }
+			get { return servant.WaitHandle; }
 		}
 
-		protected internal Language NativeLanguage {
+		internal Language NativeLanguage {
 			get {
 				check_servant ();
 				return servant.NativeLanguage;
@@ -70,7 +68,7 @@ namespace Mono.Debugger
 		//   The single-stepping engine's target state.  This will be
 		//   TargetState.RUNNING while the engine is stepping.
 		// </summary>
-		public override TargetState State {
+		public TargetState State {
 			get {
 				if (servant == null)
 					return TargetState.NoTarget;
@@ -85,7 +83,6 @@ namespace Mono.Debugger
 			lock (this) {
 				TargetEventArgs args = servant.LastTargetEvent;
 				if (args != null) {
-					is_running = false;
 					flags &= ~Flags.Background;
 					return args;
 				}
@@ -94,7 +91,7 @@ namespace Mono.Debugger
 		}
 
 		public bool IsRunning {
-			get { return is_running; }
+			get { return (servant != null) && !servant.IsStopped; }
 		}
 
 		public bool IsAlive {
@@ -107,6 +104,7 @@ namespace Mono.Debugger
 
 		public Flags ThreadFlags {
 			get { return flags; }
+			set { flags = value; }
 		}
 
 		public string Name {
@@ -132,28 +130,21 @@ namespace Mono.Debugger
 			}
 		}
 
-		internal override Architecture Architecture {
+		internal Architecture Architecture {
 			get {
 				check_servant ();
 				return servant.Architecture;
 			}
 		}
 
-		internal override ProcessServant ProcessServant {
-			get {
-				check_servant ();
-				return servant.ProcessServant;
-			}
-		}
-
 		public Process Process {
 			get {
 				check_servant ();
-				return ProcessServant.Client;
+				return servant.Process;
 			}
 		}
 
-		internal override ThreadManager ThreadManager {
+		internal ThreadManager ThreadManager {
 			get {
 				check_servant ();
 				return servant.ThreadManager;
@@ -167,11 +158,6 @@ namespace Mono.Debugger
 			}
 		}
 
-		internal void SetThreadFlags (Flags flags)
-		{
-			this.flags = flags;
-		}
-
 		public ThreadGroup ThreadGroup {
 			get {
 				check_servant ();
@@ -199,14 +185,14 @@ namespace Mono.Debugger
 		//   unnecessarily compute this several times if more than one client
 		//   accesses this property.
 		// </summary>
-		public override StackFrame CurrentFrame {
+		public StackFrame CurrentFrame {
 			get {
 				check_servant ();
 				return servant.CurrentFrame;
 			}
 		}
 
-		public override TargetAddress CurrentFrameAddress {
+		public TargetAddress CurrentFrameAddress {
 			get {
 				check_servant ();
 				return servant.CurrentFrameAddress;
@@ -222,7 +208,7 @@ namespace Mono.Debugger
 		//   without doing any stepping operations in the meantime, you'll always
 		//   get the same backtrace.
 		// </summary>
-		public override Backtrace GetBacktrace (Backtrace.Mode mode, int max_frames)
+		public Backtrace GetBacktrace (Backtrace.Mode mode, int max_frames)
 		{
 			check_servant ();
 			return servant.GetBacktrace (mode, max_frames);
@@ -243,20 +229,20 @@ namespace Mono.Debugger
 			return GetBacktrace (Backtrace.Mode.Default, -1);
 		}
 
-		public override Backtrace CurrentBacktrace {
+		public Backtrace CurrentBacktrace {
 			get {
 				check_servant ();
 				return servant.CurrentBacktrace;
 			}
 		}
 
-		public override Registers GetRegisters ()
+		public Registers GetRegisters ()
 		{
 			check_servant ();
 			return servant.GetRegisters ();
 		}
 
-		public override void SetRegisters (Registers registers)
+		public void SetRegisters (Registers registers)
 		{
 			check_alive ();
 			servant.SetRegisters (registers);
@@ -295,84 +281,78 @@ namespace Mono.Debugger
 			}
 		}
 
-		internal void SetCompleted ()
+		public CommandResult Step (ThreadingModel model, StepMode mode)
 		{
-			operation_completed_event.Set ();
+			lock (this) {
+				check_alive ();
+				return servant.Step (model, mode, null);
+			}
+		}
+
+		public CommandResult Step (ThreadingModel model, StepMode mode, StepFrame frame)
+		{
+			lock (this) {
+				check_alive ();
+				return servant.Step (model, mode, frame);
+			}
+		}
+
+		ThreadCommandResult Old_Step (StepMode mode)
+		{
+			return Old_Step (mode, null);
+		}
+
+		ThreadCommandResult Old_Step (StepMode mode, StepFrame frame)
+		{
+			lock (this) {
+				check_alive ();
+				return servant.Old_Step (mode, frame);
+			}
 		}
 
 		// <summary>
 		//   Step one machine instruction, but don't step into trampolines.
 		// </summary>
+		[Obsolete("Use Step (StepMode.SingleInstruction)")]
 		public ThreadCommandResult StepInstruction ()
 		{
-			lock (this) {
-				check_alive ();
-				is_running = true;
-				operation_completed_event.Reset ();
-				ThreadCommandResult result = new ThreadCommandResult (this);
-				servant.StepInstruction (result);
-				return result;
-			}
+			return Old_Step (StepMode.SingleInstruction);
 		}
 
 		// <summary>
 		//   Step one machine instruction, always step into method calls.
 		// </summary>
+		[Obsolete("Use Step (StepMode.NativeInstruction)")]
 		public ThreadCommandResult StepNativeInstruction ()
 		{
-			lock (this) {
-				check_alive ();
-				is_running = true;
-				operation_completed_event.Reset ();
-				ThreadCommandResult result = new ThreadCommandResult (this);
-				servant.StepNativeInstruction (result);
-				return result;
-			}
+			return Old_Step (StepMode.NativeInstruction);
 		}
 
 		// <summary>
 		//   Step one machine instruction, but step over method calls.
 		// </summary>
+		[Obsolete("Use Step (StepMode.NextInstruction)")]
 		public ThreadCommandResult NextInstruction ()
 		{
-			lock (this) {
-				check_alive ();
-				is_running = true;
-				operation_completed_event.Reset ();
-				ThreadCommandResult result = new ThreadCommandResult (this);
-				servant.NextInstruction (result);
-				return result;
-			}
+			return Old_Step (StepMode.NextInstruction);
 		}
 
 		// <summary>
 		//   Step one source line.
 		// </summary>
+		[Obsolete("Use Step (StepMode.SourceLine)")]
 		public ThreadCommandResult StepLine ()
 		{
-			lock (this) {
-				check_alive ();
-				is_running = true;
-				operation_completed_event.Reset ();
-				ThreadCommandResult result = new ThreadCommandResult (this);
-				servant.StepLine (result);
-				return result;
-			}
+			return Old_Step (StepMode.SourceLine);
 		}
 
 		// <summary>
 		//   Step one source line, but step over method calls.
 		// </summary>
+		[Obsolete("Use Step (StepMode.NextLine)")]
 		public ThreadCommandResult NextLine ()
 		{
-			lock (this) {
-				check_alive ();
-				is_running = true;
-				operation_completed_event.Reset ();
-				ThreadCommandResult result = new ThreadCommandResult (this);
-				servant.NextLine (result);
-				return result;
-			}
+			return Old_Step (StepMode.NextLine);
 		}
 
 		// <summary>
@@ -382,52 +362,52 @@ namespace Mono.Debugger
 		{
 			lock (this) {
 				check_alive ();
-				is_running = true;
-				operation_completed_event.Reset ();
-				ThreadCommandResult result = new ThreadCommandResult (this);
-				servant.Finish (native, result);
-				return result;
+
+				if (!native) {
+					if (CurrentMethod == null)
+						throw new TargetException (TargetError.NoMethod);
+
+					StepFrame step_frame = new StepFrame (
+						null, StepMode.Finish, null,
+						CurrentMethod.StartAddress, CurrentMethod.EndAddress);
+
+					return Old_Step (StepMode.Finish, step_frame);
+				} else {
+					StepFrame step_frame = new StepFrame (
+						null, StepMode.FinishNative,
+						CurrentFrame.StackPointer);
+
+					return Old_Step (StepMode.FinishNative, step_frame);
+				}
 			}
 		}
 
+		[Obsolete("Use Step (StepMode.Run)")]
 		public ThreadCommandResult Continue ()
 		{
-			return Continue (TargetAddress.Null);
+			return Old_Step (StepMode.Run);
 		}
 
+		[Obsolete("Use the new Step() API")]
 		public ThreadCommandResult Continue (TargetAddress until)
 		{
-			lock (this) {
-				check_alive ();
-				is_running = true;
-				operation_completed_event.Reset ();
-				ThreadCommandResult result = new ThreadCommandResult (this);
-				servant.Continue (until, new ThreadCommandResult (this));
-				return result;
-			}
+			return Old_Step (StepMode.Run, new StepFrame (NativeLanguage, StepMode.Run, until));
 		}
 
+		[Obsolete("Background() and Continue() are the same")]
 		public ThreadCommandResult Background ()
 		{
-			return Background (TargetAddress.Null);
+			return (ThreadCommandResult) Step (ThreadingModel.Single, StepMode.Run, null);
 		}
 
+		[Obsolete("Background() and Continue() are the same")]
 		public ThreadCommandResult Background (TargetAddress until)
 		{
-			lock (this) {
-				check_alive ();
-				is_running = true;
-				flags |= Flags.Background;
-				operation_completed_event.Reset ();
-				ThreadCommandResult result = new ThreadCommandResult (this);
-				servant.Background (until, new ThreadCommandResult (this));
-				return result;
-			}
+			return Continue (until);
 		}
 
 		internal void Kill ()
 		{
-			operation_completed_event.Set ();
 			if (servant != null)
 				servant.Kill ();
 			Dispose ();
@@ -435,7 +415,6 @@ namespace Mono.Debugger
 
 		internal void Detach ()
 		{
-			operation_completed_event.Set ();
 			if (servant != null)
 				servant.Detach ();
 			Dispose ();
@@ -451,16 +430,11 @@ namespace Mono.Debugger
 		{
 			check_alive ();
 			servant.Stop ();
-			Wait ();
-			is_running = false;
-			flags |= Flags.AutoRun;
 		}
 
-		public void Wait ()
+		public ThreadCommandResult GetWaitHandle ()
 		{
-			Report.Debug (DebugFlags.Wait, "{0} waiting", this);
-			operation_completed_event.WaitOne ();
-			Report.Debug (DebugFlags.Wait, "{0} done waiting", this);
+			return new ThreadCommandResult (this);
 		}
 
 		// <summary>
@@ -469,64 +443,22 @@ namespace Mono.Debugger
 		//   Returns a number which may be passed to RemoveBreakpoint() to remove
 		//   the breakpoint.
 		// </summary>
-		public override int InsertBreakpoint (Breakpoint breakpoint, TargetAddress address)
+		internal void InsertBreakpoint (BreakpointHandle handle,
+						TargetAddress address, int domain)
 		{
 			check_alive ();
-			return servant.InsertBreakpoint (breakpoint, address);
+			servant.InsertBreakpoint (handle, address, domain);
 		}
 
 		// <summary>
 		//   Remove breakpoint @index.  @index is the breakpoint number which has
 		//   been returned by InsertBreakpoint().
 		// </summary>
-		public void RemoveBreakpoint (int index)
+		internal void RemoveBreakpoint (BreakpointHandle handle)
 		{
 			check_disposed ();
 			if (servant != null)
-				servant.RemoveBreakpoint (index);
-		}
-
-		// <summary>
-		//   Insert a breakpoint at function @func.
-		//
-		//   Returns a number which may be passed to RemoveBreakpoint() to remove
-		//   the breakpoint.
-		// </summary>
-		public int InsertBreakpoint (Breakpoint breakpoint, TargetFunctionType func)
-		{
-			CommandResult result;
-
-			lock (this) {
-				check_alive ();
-				result = servant.InsertBreakpoint (breakpoint, func);
-			}
-
-			result.Wait ();
-
-			return (int) result.Result;
-		}
-
-		// <summary>
-		//   Add an event handler.
-		//
-		//   Returns a number which may be passed to RemoveEventHandler() to remove
-		//   the event handler.
-		// </summary>
-		public int AddEventHandler (Event handle)
-		{
-			check_alive ();
-			return servant.AddEventHandler (handle);
-		}
-
-		// <summary>
-		//   Remove event handler @index.  @index is the event handler number which has
-		//   been returned by AddEventHandler().
-		// </summary>
-		public void RemoveEventHandler (int index)
-		{
-			check_disposed ();
-			if (servant != null)
-				servant.RemoveEventHandler (index);
+				servant.RemoveBreakpoint (handle);
 		}
 
 		public string PrintObject (Style style, TargetObject obj, DisplayFormat format)
@@ -541,45 +473,75 @@ namespace Mono.Debugger
 			return servant.PrintType (style, type);
 		}
 
+#region User Threads
+
+		void IOperationHost.OperationCompleted (SingleSteppingEngine sse, TargetEventArgs args, ThreadingModel model)
+		{ }
+
+		void IOperationHost.SendResult (SingleSteppingEngine sse, TargetEventArgs args)
+		{
+			sse.Process.Debugger.OnTargetEvent (sse.Client, args);
+		}
+
+		void IOperationHost.Abort ()
+		{
+			Stop ();
+		}
+
+#endregion
+
 		//
 		// Disassembling.
 		//
 
-		public override int GetInstructionSize (TargetAddress address)
+		public int GetInstructionSize (TargetAddress address)
 		{
 			check_alive ();
 			return servant.GetInstructionSize (address);
 		}
 
-		public override AssemblerLine DisassembleInstruction (Method method, TargetAddress address)
+		public AssemblerLine DisassembleInstruction (Method method, TargetAddress address)
 		{
 			check_alive ();
 			return servant.DisassembleInstruction (method, address);
 		}
 
-		public override AssemblerMethod DisassembleMethod (Method method)
+		public AssemblerMethod DisassembleMethod (Method method)
 		{
 			check_alive ();
 			return servant.DisassembleMethod (method);
 		}
 
+		[Obsolete]
 		public RuntimeInvokeResult RuntimeInvoke (TargetFunctionType function,
-							  TargetClassObject object_argument,
+							  TargetStructObject object_argument,
 							  TargetObject[] param_objects,
 							  bool is_virtual, bool debug)
+		{
+			RuntimeInvokeFlags flags = RuntimeInvokeFlags.None;
+			if (is_virtual)
+				flags |= RuntimeInvokeFlags.VirtualMethod;
+			if (debug)
+				flags |= RuntimeInvokeFlags.BreakOnEntry;
+			return RuntimeInvoke (function, object_argument, param_objects, flags);
+		}
+
+		public RuntimeInvokeResult RuntimeInvoke (TargetFunctionType function,
+							  TargetStructObject object_argument,
+							  TargetObject[] param_objects,
+							  RuntimeInvokeFlags flags)
 		{
 			lock (this) {
 				check_alive ();
 				RuntimeInvokeResult result = new RuntimeInvokeResult (this);
 				servant.RuntimeInvoke (
-					function, object_argument, param_objects, is_virtual,
-					debug, result);
+					function, object_argument, param_objects,
+					flags, result);
 				return result;
 			}
 		}
 
-		public TargetAddress CallMethod (TargetAddress method, TargetAddress arg1,
-						 TargetAddress arg2)
+		public TargetAddress CallMethod (TargetAddress method, long arg1, long arg2)
 		{
 			CommandResult result;
 
@@ -596,14 +558,14 @@ namespace Mono.Debugger
 			return (TargetAddress) result.Result;
 		}
 
-		public TargetAddress CallMethod (TargetAddress method, long method_arg,
-						 string string_arg)
+		public TargetAddress CallMethod (TargetAddress method, TargetAddress arg1,
+						 long arg2)
 		{
 			CommandResult result;
 
 			lock (this) {
 				check_alive ();
-				result = servant.CallMethod (method, method_arg, string_arg);
+				result = servant.CallMethod (method, arg1.Address, arg2);
 			}
 
 			result.Wait ();
@@ -614,13 +576,51 @@ namespace Mono.Debugger
 			return (TargetAddress) result.Result;
 		}
 
-		public void Return (bool run_finally)
+		public TargetAddress CallMethod (TargetAddress method, TargetAddress arg1,
+						 long arg2, long arg3, string string_arg)
 		{
 			CommandResult result;
 
 			lock (this) {
 				check_alive ();
-				result = servant.Return (run_finally);
+				result = servant.CallMethod (
+					method, arg1.Address, arg2, arg3, string_arg);
+			}
+
+			result.Wait ();
+
+			if (result.Result == null)
+				throw new TargetException (TargetError.UnknownError);
+
+			return (TargetAddress) result.Result;
+		}
+
+		public TargetAddress CallMethod (TargetAddress method, TargetAddress method_argument,
+						 TargetObject object_argument)
+		{
+			CommandResult result;
+
+			lock (this) {
+				check_alive ();
+				result = servant.CallMethod (
+					method, method_argument, object_argument);
+			}
+
+			result.Wait ();
+
+			if (result.Result == null)
+				throw new TargetException (TargetError.UnknownError);
+
+			return (TargetAddress) result.Result;
+		}
+
+		public void Return (ReturnMode mode)
+		{
+			CommandResult result;
+
+			lock (this) {
+				check_alive ();
+				result = servant.Return (mode);
 				if (result == null)
 					return;
 			}
@@ -628,16 +628,12 @@ namespace Mono.Debugger
 			result.Wait ();
 		}
 
-		public void AbortInvocation ()
+		internal void AbortInvocation (long rti_id)
 		{
-			CommandResult result;
-
 			lock (this) {
 				check_alive ();
-				result = servant.AbortInvocation ();
+				servant.AbortInvocation (rti_id);
 			}
-
-			result.Wait ();
 		}
 
 		public string PrintRegisters (StackFrame frame)
@@ -661,28 +657,28 @@ namespace Mono.Debugger
 			get { return (servant != null) && servant.IsStopped; }
 		}
 
-		public override TargetInfo TargetInfo {
+		public TargetMemoryInfo TargetMemoryInfo {
 			get {
 				check_servant ();
-				return servant.TargetInfo;
+				return servant.TargetMemoryInfo;
 			}
 		}
 
 #region ITargetInfo implementation
-		public override int TargetAddressSize {
-			get { return TargetInfo.TargetAddressSize; }
+		public int TargetAddressSize {
+			get { return TargetMemoryInfo.TargetAddressSize; }
 		}
 
-		public override int TargetIntegerSize {
-			get { return TargetInfo.TargetIntegerSize; }
+		public int TargetIntegerSize {
+			get { return TargetMemoryInfo.TargetIntegerSize; }
 		}
 
-		public override int TargetLongIntegerSize {
-			get { return TargetInfo.TargetLongIntegerSize; }
+		public int TargetLongIntegerSize {
+			get { return TargetMemoryInfo.TargetLongIntegerSize; }
 		}
 
-		public override bool IsBigEndian {
-			get { return TargetInfo.IsBigEndian; }
+		public bool IsBigEndian {
+			get { return TargetMemoryInfo.IsBigEndian; }
 		}
 #endregion
 
@@ -693,87 +689,87 @@ namespace Mono.Debugger
 			servant.WriteBuffer (address, buffer);
 		}
 
-		public override AddressDomain AddressDomain {
+		public AddressDomain AddressDomain {
 			get {
-				return TargetInfo.AddressDomain;
+				return TargetMemoryInfo.AddressDomain;
 			}
 		}
 
-		public override byte ReadByte (TargetAddress address)
+		public byte ReadByte (TargetAddress address)
 		{
 			check_alive ();
 			return servant.ReadByte (address);
 		}
 
-		public override int ReadInteger (TargetAddress address)
+		public int ReadInteger (TargetAddress address)
 		{
 			check_alive ();
 			return servant.ReadInteger (address);
 		}
 
-		public override long ReadLongInteger (TargetAddress address)
+		public long ReadLongInteger (TargetAddress address)
 		{
 			check_alive ();
 			return servant.ReadLongInteger (address);
 		}
 
-		public override TargetAddress ReadAddress (TargetAddress address)
+		public TargetAddress ReadAddress (TargetAddress address)
 		{
 			check_alive ();
 			return servant.ReadAddress (address);
 		}
 
-		public override string ReadString (TargetAddress address)
+		public string ReadString (TargetAddress address)
 		{
 			check_alive ();
 			return servant.ReadString (address);
 		}
 
-		public override TargetBlob ReadMemory (TargetAddress address, int size)
+		public TargetBlob ReadMemory (TargetAddress address, int size)
 		{
 			check_alive ();
 			byte[] buffer = servant.ReadBuffer (address, size);
-			return new TargetBlob (buffer, TargetInfo);
+			return new TargetBlob (buffer, TargetMemoryInfo);
 		}
 
-		public override byte[] ReadBuffer (TargetAddress address, int size)
+		public byte[] ReadBuffer (TargetAddress address, int size)
 		{
 			check_alive ();
 			return servant.ReadBuffer (address, size);
 		}
 
-		public override bool CanWrite {
+		public bool CanWrite {
 			get {
 				check_servant ();
 				return servant.CanWrite;
 			}
 		}
 
-		public override void WriteBuffer (TargetAddress address, byte[] buffer)
+		public void WriteBuffer (TargetAddress address, byte[] buffer)
 		{
 			write_memory (address, buffer);
 		}
 
-		public override void WriteByte (TargetAddress address, byte value)
+		public void WriteByte (TargetAddress address, byte value)
 		{
 			throw new InvalidOperationException ();
 		}
 
-		public override void WriteInteger (TargetAddress address, int value)
+		public void WriteInteger (TargetAddress address, int value)
 		{
 			throw new InvalidOperationException ();
 		}
 
-		public override void WriteLongInteger (TargetAddress address, long value)
+		public void WriteLongInteger (TargetAddress address, long value)
 		{
 			throw new InvalidOperationException ();
 		}
 
-		public override void WriteAddress (TargetAddress address, TargetAddress value)
+		public void WriteAddress (TargetAddress address, TargetAddress value)
 		{
 			check_alive ();
 			TargetBinaryWriter writer = new TargetBinaryWriter (
-				TargetInfo.TargetAddressSize, TargetInfo);
+				TargetMemoryInfo.TargetAddressSize, TargetMemoryInfo);
 			writer.WriteAddress (value);
 			write_memory (address, writer.Contents);
 		}
@@ -782,23 +778,13 @@ namespace Mono.Debugger
 #region IDisposable implementation
 		private bool disposed = false;
 
-		protected void check_disposed ()
+		void check_disposed ()
 		{
 			if (disposed)
 				throw new ObjectDisposedException ("Thread");
 		}
 
-		protected virtual void DoDispose ()
-		{
-			if (servant != null) {
-				servant.Dispose ();
-				servant = null;
-
-				operation_completed_event.Set ();
-			}
-		}
-
-		protected virtual void Dispose (bool disposing)
+		void Dispose (bool disposing)
 		{
 			// Check to see if Dispose has already been called.
 			if (disposed)
@@ -808,7 +794,10 @@ namespace Mono.Debugger
 
 			// If this is a call to Dispose, dispose all managed resources.
 			if (disposing) {
-				DoDispose ();
+				if (servant != null) {
+					servant.Dispose ();
+					servant = null;
+				}
 			}
 		}
 
@@ -826,80 +815,81 @@ namespace Mono.Debugger
 		}
 	}
 
-	public abstract class CommandResult : DebuggerMarshalByRefObject
+	public class ThreadCommandResult : OperationCommandResult
 	{
-		public object Result;
-
-		public abstract ST.WaitHandle CompletedEvent {
-			get;
-		}
-
-		public abstract void Completed ();
-
-		public abstract void Abort ();
-
-		public void Wait ()
-		{
-			CompletedEvent.WaitOne ();
-			if (Result is Exception)
-				throw (Exception) Result;
-		}
-	}
-
-	public class ThreadCommandResult : CommandResult
-	{
-		Thread thread;
-
-		internal ThreadCommandResult (Thread thread)
-		{
-			this.thread = thread;
-		}
+		protected ST.ManualResetEvent completed_event = new ST.ManualResetEvent (false);
 
 		public Thread Thread {
-			get { return thread; }
+			get; private set;
 		}
 
-		public override ST.WaitHandle CompletedEvent {
-			get { return thread.WaitHandle; }
+		internal override IOperationHost Host {
+			get { return Thread; }
 		}
 
-		public override void Abort ()
+		internal ThreadCommandResult (Thread thread)
+			: base (ThreadingModel.Single)
 		{
-			thread.Stop ();
-		}
-
-		public override void Completed ()
-		{
-			thread.SetCompleted ();
-		}
-	}
-
-	public class RuntimeInvokeResult : CommandResult
-	{
-		Thread thread;
-		ST.ManualResetEvent completed_event = new ST.ManualResetEvent (false);
-
-		internal RuntimeInvokeResult (Thread thread)
-		{
-			this.thread = thread;
+			this.Thread = thread;
 		}
 
 		public override ST.WaitHandle CompletedEvent {
 			get { return completed_event; }
 		}
 
-		public override void Completed ()
+		internal override void Completed ()
 		{
 			completed_event.Set ();
 		}
 
+		internal override void OnExecd (SingleSteppingEngine new_thread)
+		{
+			Thread = new_thread.Thread;
+		}
+	}
+
+	public class RuntimeInvokeResult : ThreadCommandResult
+	{
+		internal RuntimeInvokeResult (Thread thread)
+			: base (thread)
+		{ }
+
 		public override void Abort ()
 		{
-			thread.Stop ();
+			Thread.AbortInvocation (ID);
+			completed_event.WaitOne ();
 		}
 
+		internal override void Completed (SingleSteppingEngine sse, TargetEventArgs args)
+		{
+			Host.OperationCompleted (sse, args, ThreadingModel);
+			if (args != null)
+				Host.SendResult (sse, args);
+		}
+
+		public long ID;
+		public bool InvocationAborted;
 		public bool InvocationCompleted;
 		public TargetObject ReturnObject;
 		public string ExceptionMessage;
+		public TargetException TargetException;
+	}
+
+	[Flags]
+	public enum RuntimeInvokeFlags
+	{
+		None			= 0,
+		NoSideEffects		= 1,
+		NestedBreakStates	= 2,
+		BreakOnEntry		= 4,
+		VirtualMethod		= 8,
+		SendEventOnCompletion	= 16
+	}
+
+	public enum ReturnMode
+	{
+		Managed,
+		Native,
+		Invocation
 	}
 }
