@@ -4,158 +4,13 @@ using System.IO;
 using System.Reflection;
 using System.Collections;
 using System.Globalization;
+using ST=System.Threading;
 using Mono.Debugger;
 using Mono.Debugger.Languages;
 using Mono.Debugger.Languages.Mono;
 
 namespace Mono.Debugger.Frontend
 {
-	public class ExpressionParser : IExpressionParser
-	{
-		public readonly Interpreter Interpreter;
-
-		private readonly CSharp.ExpressionParser parser;
-
-		internal ExpressionParser (Interpreter interpreter)
-		{
-			this.Interpreter = interpreter;
-
-			parser = new CSharp.ExpressionParser ("C#");
-		}
-
-		public Expression Parse (string expression)
-		{
-			return parser.Parse (expression);
-		}
-
-		protected static SourceLocation FindFile (ScriptingContext context, string filename,
-							  int line)
-		{
-			SourceFile file = context.Interpreter.Session.FindFile (filename);
-			if (file == null)
-				throw new ScriptingException ("Cannot find source file `{0}'.",
-							      filename);
-
-			MethodSource source = file.FindMethod (line);
-			if (source == null)
-				throw new ScriptingException (
-					"Cannot find method corresponding to line {0} in `{1}'.",
-					line, file.Name);
-
-			return new SourceLocation (source, file, line);
-		}
-
-		protected SourceLocation DoParseExpression (ScriptingContext context,
-							    LocationType type, string arg)
-		{
-			Expression expr = Interpreter.ExpressionParser.Parse (arg);
-			if (expr == null)
-				throw new ScriptingException ("Cannot resolve expression `{0}'.", arg);
-
-			MethodExpression mexpr = expr.ResolveMethod (context, type);
-
-			if (mexpr != null)
-				return mexpr.EvaluateSource (context);
-			else
-				return context.FindMethod (arg);
-		}
-
-		public static bool ParseLocation (ScriptingContext context, string arg,
-						  out SourceLocation location)
-		{
-			int line;
-			int pos = arg.IndexOf (':');
-			if (pos >= 0) {
-				string filename = arg.Substring (0, pos);
-				try {
-					line = (int) UInt32.Parse (arg.Substring (pos+1));
-				} catch {
-					throw new ScriptingException ("Expected filename:line");
-				}
-
-				location = FindFile (context, filename, line);
-				return true;
-			}
-
-			try {
-				line = (int) UInt32.Parse (arg);
-			} catch {
-				location = null;
-				return false;
-			}
-
-			StackFrame frame = context.CurrentFrame;
-			if ((frame == null) || (frame.SourceLocation == null) ||
-			    (frame.SourceLocation.FileName == null))
-				throw new ScriptingException (
-					"Current stack frame doesn't have source code");
-
-			location = FindFile (context, frame.SourceLocation.FileName, line);
-			return true;
-		}
-
-		protected SourceLocation DoParse (ScriptingContext context, LocationType type,
-						  string arg)
-		{
-			if (type != LocationType.Default)
-				return DoParseExpression (context, type, arg);
-
-			SourceLocation location;
-			if (ParseLocation (context, arg, out location))
-				return location;
-
-			return DoParseExpression (context, type, arg);
-		}
-
-		public SourceLocation Parse (Thread target, StackFrame frame,
-					     LocationType type, string arg)
-		{
-			ScriptingContext context = new ScriptingContext (Interpreter);
-			context.CurrentThread = target;
-			context.CurrentFrame = frame;
-
-			try {
-				return DoParse (context, type, arg);
-			} catch (ScriptingException ex) {
-				throw new TargetException (TargetError.LocationInvalid, ex.Message);
-			}
-		}
-
-		public string EvaluateExpression (ScriptingContext context, string text,
-						  DisplayFormat format)
-		{
-			Expression expression;
-
-			try {
-				expression = Interpreter.ExpressionParser.Parse (text);
-			} catch (ScriptingException ex) {
-				throw new ScriptingException ("Cannot parse expression `{0}': {1}.",
-							      text, ex.Message);
-			} catch {
-				throw new ScriptingException ("Cannot parse expression `{0}'.", text);
-			}
-
-			try {
-				expression = expression.Resolve (context);
-			} catch (ScriptingException ex) {
-				throw new ScriptingException ("Cannot resolve expression `{0}': {1}",
-							      text, ex.Message);
-			} catch {
-				throw new ScriptingException ("Cannot resolve expression `{0}'.", text);
-			}
-
-			try {
-				object retval = expression.Evaluate (context);
-				return context.FormatObject (retval, format);
-			} catch (ScriptingException ex) {
-				throw new ScriptingException ("Cannot evaluate expression `{0}': {1}",
-							      text, ex.Message);
-			} catch {
-				throw new ScriptingException ("Cannot evaluate expression `{0}'.", text);
-			}
-		}
-	}
-
 	public abstract class Expression
 	{
 		public abstract string Name {
@@ -250,7 +105,7 @@ namespace Mono.Debugger.Frontend
 			if (obj == null)
 				return null;
 
-			TargetStructObject cobj = obj as TargetStructObject;
+			TargetClassObject cobj = obj as TargetClassObject;
 			if (cobj != null) {
 				TargetObject current;
 				try {
@@ -483,14 +338,13 @@ namespace Mono.Debugger.Frontend
 
 		protected override TargetObject DoEvaluateObject (ScriptingContext context)
 		{
-			StackFrame frame = context.CurrentFrame;
-			if ((frame.Language == null) ||
-			    !frame.Language.CanCreateInstance (val.GetType ()))
+			if ((context.CurrentLanguage == null) ||
+			    !context.CurrentLanguage.CanCreateInstance (val.GetType ()))
 				throw new ScriptingException (
 					"Cannot instantiate value '{0}' in the current frame's " +
 					"language", Name);
 
-			return frame.Language.CreateInstance (frame.Thread, val);
+			return context.CurrentLanguage.CreateInstance (context.CurrentThread, val);
 		}
 
 		public override TargetAddress EvaluateAddress (ScriptingContext context)
@@ -535,12 +389,11 @@ namespace Mono.Debugger.Frontend
 
 		protected override TargetObject DoEvaluateObject (ScriptingContext context)
 		{
-			StackFrame frame = context.CurrentFrame;
-			if ((frame.Language == null) ||
-			    !frame.Language.CanCreateInstance (typeof (string)))
+			if ((context.CurrentLanguage == null) ||
+			    (!context.CurrentLanguage.CanCreateInstance (typeof (string))))
 				throw new ScriptingException ("Cannot instantiate value '{0}' in the current frame's language", Name);
 
-			return frame.Language.CreateInstance (frame.Thread, val);
+			return context.CurrentLanguage.CreateInstance (context.CurrentThread, val);
 		}
 
 		public override string ToString ()
@@ -575,12 +428,11 @@ namespace Mono.Debugger.Frontend
 
 		protected override TargetObject DoEvaluateObject (ScriptingContext context)
 		{
-			StackFrame frame = context.CurrentFrame;
-			if ((frame.Language == null) ||
-			    !frame.Language.CanCreateInstance (typeof (bool)))
+			if ((context.CurrentLanguage == null) ||
+			    !context.CurrentLanguage.CanCreateInstance (typeof (bool)))
 				throw new ScriptingException ("Cannot instantiate value '{0}' in the current frame's language", Name);
 
-			return frame.Language.CreateInstance (frame.Thread, val);
+			return context.CurrentLanguage.CreateInstance (context.CurrentThread, val);
 		}
 
 		public override string ToString ()
@@ -911,14 +763,20 @@ namespace Mono.Debugger.Frontend
 			if (method == null)
 				return null;
 
-			TargetStructType decl_type = method.GetDeclaringType (context.CurrentThread);
+			TargetClassType decl_type = method.GetDeclaringType (context.CurrentThread);
 			if (decl_type == null)
 				return null;
 
-			TargetStructObject instance = null;
+			TargetClassObject instance = null;
 			if (method.HasThis) {
 				TargetVariable this_var = method.GetThis (context.CurrentThread);
-				instance = (TargetStructObject) this_var.GetObject (frame);
+				TargetObject this_obj = this_var.GetObject (frame);
+
+				var pobj = this_obj as TargetPointerObject;
+				if (pobj != null)
+					this_obj = pobj.GetDereferencedObject (context.CurrentThread);
+
+				instance = (TargetClassObject) this_obj;
 			}
 
 			member = StructAccessExpression.FindMember (
@@ -935,7 +793,7 @@ namespace Mono.Debugger.Frontend
 			if (member != null)
 				return member;
 
-			string[] namespaces = context.GetNamespaces (frame);
+			string[] namespaces = context.GetNamespaces ();
 			if (namespaces == null)
 				return null;
 
@@ -966,6 +824,22 @@ namespace Mono.Debugger.Frontend
 				TargetAddress address = context.CurrentProcess.LookupSymbol (name);
 				if (!address.IsNull)
 					return new NumberExpression (address.Address);
+			} else if (context.ImplicitInstance != null) {
+				MemberExpression member = StructAccessExpression.FindMember (
+					context.CurrentThread, context.ImplicitInstance.Type,
+					context.ImplicitInstance, name, true, true);
+				if (member != null)
+					return member;
+
+				string[] namespaces = context.GetNamespaces () ?? new string [0];
+				foreach (string ns in namespaces) {
+					string full_name = MakeFQN (ns, name);
+					member = StructAccessExpression.FindMember (
+						context.CurrentThread, context.ImplicitInstance.Type,
+						context.ImplicitInstance, full_name, true, true);
+					if (member != null)
+						return member;
+				}
 			}
 
 			SourceLocation location = context.FindMethod (name);
@@ -1011,7 +885,7 @@ namespace Mono.Debugger.Frontend
 			if (type != null)
 				return new TypeExpression (type);
 
-			string[] namespaces = context.GetNamespaces (context.CurrentFrame);
+			string[] namespaces = context.GetNamespaces ();
 			if (namespaces == null)
 				return null;
 
@@ -1135,7 +1009,7 @@ namespace Mono.Debugger.Frontend
 			}
 
 			if (lexpr != null) {
-				TargetStructObject sobj = Convert.ToStructObject (
+				TargetClassObject sobj = Convert.ToStructObject (
 					target, lexpr.EvaluateObject (context));
 				if (sobj == null)
 					throw new ScriptingException (
@@ -1221,16 +1095,14 @@ namespace Mono.Debugger.Frontend
 
 		protected override Expression DoResolveType (ScriptingContext context)
 		{
-			StackFrame frame = context.CurrentFrame;
-
 			TargetType the_type;
 
 			Expression ltype = left.TryResolveType (context);
 			if (ltype == null)
-				the_type = frame.Language.LookupType (Name);
+				the_type = context.CurrentLanguage.LookupType (Name);
 			else {
 				string nested = ltype.Name + "+" + name;
-				the_type = frame.Language.LookupType (nested);
+				the_type = context.CurrentLanguage.LookupType (nested);
 			}
 
 			if (the_type == null)
@@ -1694,9 +1566,9 @@ namespace Mono.Debugger.Frontend
 		protected readonly TargetStructObject instance;
 		TargetClass class_info;
 
-		protected StructAccessExpression (TargetStructType type,
-						  TargetStructObject instance,
-						  TargetMemberInfo member)
+		public StructAccessExpression (TargetStructType type,
+					       TargetStructObject instance,
+					       TargetMemberInfo member)
 		{
 			this.Type = type;
 			this.Member = member;
@@ -1705,7 +1577,7 @@ namespace Mono.Debugger.Frontend
 		}
 
 		public override string Name {
-			get { return Member.Name; }
+			get { return Type.Name + "." + Member.Name; }
 		}
 
 		public override TargetStructObject InstanceObject {
@@ -1738,7 +1610,7 @@ namespace Mono.Debugger.Frontend
 			if (class_info != null)
 				return;
 
-			class_info = Type.GetClass (target);
+			class_info = Type.ForceClassInitialization (target);
 			if (class_info == null)
 				throw new ScriptingException ("Class `{0}' not initialized yet.",
 							      Type.Name);
@@ -1799,9 +1671,11 @@ namespace Mono.Debugger.Frontend
 		protected TargetObject GetProperty (ScriptingContext context,
 						    TargetPropertyInfo prop)
 		{
-			RuntimeInvokeResult result = context.Interpreter.RuntimeInvoke (
+			RuntimeInvokeFlags flags = context.GetRuntimeInvokeFlags ();
+
+			RuntimeInvokeResult result = context.RuntimeInvoke (
 				context.CurrentThread, prop.Getter, InstanceObject,
-				new TargetObject [0], true, false);
+				new TargetObject [0], flags);
 
 			if (result.ExceptionMessage != null)
 				throw new ScriptingException (
@@ -1931,14 +1805,12 @@ namespace Mono.Debugger.Frontend
 
 		protected override TargetObject DoEvaluateObject (ScriptingContext context)
 		{
-			StackFrame frame = context.CurrentFrame;
-
 			if (!Member.IsStatic && (InstanceObject == null))
 				throw new ScriptingException (
 					"Instance member `{0}' cannot be used in static context.", Name);
 
 			try {
-				return GetMember (context, frame.Thread, Member);
+				return GetMember (context, context.CurrentThread, Member);
 			} catch (TargetException ex) {
 				throw new ScriptingException ("Cannot access struct member `{0}': {1}",
 							      Name, ex.Message);
@@ -2021,9 +1893,11 @@ namespace Mono.Debugger.Frontend
 			if (prop.Setter == null)
 				throw new ScriptingException ("Property `{0}' has no setter.", Name);
 
-			RuntimeInvokeResult result = context.Interpreter.RuntimeInvoke (
+			RuntimeInvokeFlags flags = context.GetRuntimeInvokeFlags ();
+
+			RuntimeInvokeResult result = context.RuntimeInvoke (
 				context.CurrentThread, prop.Setter, InstanceObject,
-				new TargetObject [] { obj }, true, false);
+				new TargetObject [] { obj }, flags);
 
 			if (result.ExceptionMessage != null)
 				throw new ScriptingException (
@@ -2034,20 +1908,10 @@ namespace Mono.Debugger.Frontend
 		protected override bool DoAssign (ScriptingContext context, TargetObject obj)
 		{
 			if (Member is TargetFieldInfo) {
-				if (Member.Type != obj.Type)
-					throw new ScriptingException (
-						"Type mismatch: cannot assign expression of type " +
-						"`{0}' to field `{1}', which is of type `{2}'.",
-						obj.TypeName, Name, Member.Type.Name);
-
+				obj = Convert.ImplicitConversionRequired (context, obj, Member.Type);
 				SetField (context.CurrentThread, (TargetFieldInfo) Member, obj);
 			} else if (Member is TargetPropertyInfo) {
-				if (Member.Type != obj.Type)
-					throw new ScriptingException (
-						"Type mismatch: cannot assign expression of type " +
-						"`{0}' to property `{1}', which is of type `{2}'.",
-						obj.TypeName, Name, Member.Type.Name);
-
+				obj = Convert.ImplicitConversionRequired (context, obj, Member.Type);
 				SetProperty (context, (TargetPropertyInfo) Member, obj);
 			} else if (Member is TargetEventInfo)
 				throw new ScriptingException ("Can't set events directly.");
@@ -2465,10 +2329,10 @@ namespace Mono.Debugger.Frontend
 			return this;
 		}
 
-		static TargetStructObject TryParentCast (ScriptingContext context,
-							 TargetStructObject source,
-							 TargetStructType source_type,
-							 TargetStructType target_type)
+		static TargetClassObject TryParentCast (ScriptingContext context,
+							 TargetClassObject source,
+							 TargetClassType source_type,
+							 TargetClassType target_type)
 		{
 			if (source_type == target_type)
 				return source;
@@ -2476,7 +2340,7 @@ namespace Mono.Debugger.Frontend
 			if (!source_type.HasParent)
 				return null;
 
-			TargetStructType parent_type = source_type.GetParentType (context.CurrentThread);
+			TargetClassType parent_type = source_type.GetParentType (context.CurrentThread);
 			source = TryParentCast (context, source, parent_type, target_type);
 			if (source == null)
 				return null;
@@ -2484,11 +2348,11 @@ namespace Mono.Debugger.Frontend
 			return source.GetParentObject (context.CurrentThread) as TargetClassObject;
 		}
 
-		static TargetStructObject TryCurrentCast (ScriptingContext context,
+		static TargetClassObject TryCurrentCast (ScriptingContext context,
 							  TargetClassObject source,
 							  TargetClassType target_type)
 		{
-			TargetStructObject current = source.GetCurrentObject (context.CurrentThread);
+			TargetClassObject current = source.GetCurrentObject (context.CurrentThread);
 			if (current == null)
 				return null;
 
@@ -2505,7 +2369,7 @@ namespace Mono.Debugger.Frontend
 			if (sobj == null)
 				return null;
 
-			TargetStructObject result = TryParentCast (context, sobj, sobj.Type, target_type);
+			TargetClassObject result = TryParentCast (context, sobj, sobj.Type, target_type);
 			if (result != null)
 				return result;
 
@@ -2513,8 +2377,8 @@ namespace Mono.Debugger.Frontend
 		}
 
 		static bool TryParentCast (ScriptingContext context,
-					   TargetStructType source_type,
-					   TargetStructType target_type)
+					   TargetClassType source_type,
+					   TargetClassType target_type)
 		{
 			if (source_type == target_type)
 				return true;
@@ -2522,7 +2386,7 @@ namespace Mono.Debugger.Frontend
 			if (!source_type.HasParent)
 				return false;
 
-			TargetStructType parent_type = source_type.GetParentType (context.CurrentThread);
+			TargetClassType parent_type = source_type.GetParentType (context.CurrentThread);
 			return TryParentCast (context, parent_type, target_type);
 		}
 
@@ -2563,6 +2427,9 @@ namespace Mono.Debugger.Frontend
 				    (source is TargetFundamentalObject))
 					return target_type.Language.CreateBoxedObject (context.CurrentThread, source);
 
+				if (source is TargetObjectObject)
+					return source;
+
 				throw new ScriptingException (
 					"Cannot box object `{0}': not a value-type", expr.Name);
 			}
@@ -2705,6 +2572,11 @@ namespace Mono.Debugger.Frontend
 					return true;
 				break;
 
+			case FundamentalKind.Decimal:
+				if (tkind == FundamentalKind.Decimal)
+					return true;
+				break;
+
 			default:
 				break;
 			}
@@ -2745,6 +2617,8 @@ namespace Mono.Debugger.Frontend
 				return System.Convert.ToSingle (value);
 			case FundamentalKind.Double:
 				return System.Convert.ToDouble (value);
+			case FundamentalKind.Decimal:
+				return System.Convert.ToDecimal (value);
 			default:
 				return null;
 			}
@@ -2792,8 +2666,8 @@ namespace Mono.Debugger.Frontend
 		}
 
 		static bool ImplicitReferenceConversionExists (ScriptingContext context,
-							       TargetStructType source,
-							       TargetStructType target)
+							       TargetClassType source,
+							       TargetClassType target)
 		{
 			if (source == target)
 				return true;
@@ -2801,7 +2675,7 @@ namespace Mono.Debugger.Frontend
 			if (!source.HasParent)
 				return false;
 
-			TargetStructType parent_type = source.GetParentType (context.CurrentThread);
+			TargetClassType parent_type = source.GetParentType (context.CurrentThread);
 			return ImplicitReferenceConversionExists (context, parent_type, target);
 		}
 
@@ -2829,6 +2703,10 @@ namespace Mono.Debugger.Frontend
 					context, (TargetFundamentalType) source,
 					(TargetFundamentalType) target);
 
+			if ((source is TargetFundamentalType) && (target is TargetObjectType) &&
+			    (((TargetFundamentalType) source).FundamentalKind == FundamentalKind.String))
+				return true;
+
 			if ((source is TargetClassType) && (target is TargetClassType))
 				return ImplicitReferenceConversionExists (
 					context, (TargetClassType) source,
@@ -2848,11 +2726,27 @@ namespace Mono.Debugger.Frontend
 					context, (TargetFundamentalObject) obj,
 					(TargetFundamentalType) type);
 
+			if ((obj is TargetFundamentalObject) && (type is TargetObjectType) &&
+			    (((TargetFundamentalType) obj.Type).FundamentalKind == FundamentalKind.String))
+				return obj.Type.Language.CreateBoxedObject (context.CurrentThread, obj);
+
 			if ((obj is TargetClassObject) && (type is TargetClassType))
 				return ImplicitReferenceConversion (
 					context, (TargetClassObject) obj,
 					(TargetClassType) type);
 
+			TargetNullableType ntype = type as TargetNullableType;
+			if (ntype != null) {
+				if (obj.Kind == TargetObjectKind.Null)
+					return obj;
+				else if (obj.Kind != TargetObjectKind.Nullable)
+					return ImplicitConversion (context, obj, ntype.ElementType);
+
+				TargetNullableType ntype2 = (TargetNullableType) obj.Type;
+				if (ImplicitConversionExists (context, ntype2.ElementType, ntype.ElementType))
+					return obj;
+			}
+
 			return null;
 		}
 
@@ -2907,9 +2801,9 @@ namespace Mono.Debugger.Frontend
 			return null;
 		}
 
-		public static TargetStructObject ToStructObject (Thread target, TargetObject obj)
+		public static TargetClassObject ToStructObject (Thread target, TargetObject obj)
 		{
-			TargetStructObject sobj = obj as TargetStructObject;
+			TargetClassObject sobj = obj as TargetClassObject;
 			if (sobj != null)
 				return sobj;
 
@@ -3164,22 +3058,26 @@ namespace Mono.Debugger.Frontend
 				Thread thread = context.CurrentThread;
 				RuntimeInvokeResult result;
 
-				result = context.Interpreter.RuntimeInvoke (
-					thread, method, instance, objs, true, debug);
+				RuntimeInvokeFlags flags = context.GetRuntimeInvokeFlags ();
+				if (debug)
+					flags |= RuntimeInvokeFlags.BreakOnEntry | RuntimeInvokeFlags.SendEventOnCompletion |
+						RuntimeInvokeFlags.NestedBreakStates;
+
+				result = context.RuntimeInvoke (thread, method, instance, objs, flags);
 
 				if (result == null)
 					throw new ScriptingException (
 						"Invocation of `{0}' aborted abnormally.", Name);
 
 				if (result.ExceptionMessage != null)
-					throw new ScriptingException (
-						"Invocation of `{0}' raised an exception: {1}",
-						Name, result.ExceptionMessage);
+					throw new InvocationException (Name, result.ExceptionMessage, result.ReturnObject);
 
 				return result.ReturnObject;
 			} catch (TargetException ex) {
 				throw new ScriptingException (
 					"Invocation of `{0}' raised an exception: {1}", Name, ex.Message);
+			} catch (EvaluationTimeoutException ex) {
+				throw new ScriptingException ("Invocation of `{0}' timed out.", Name);
 			}
 		}
 
@@ -3290,7 +3188,7 @@ namespace Mono.Debugger.Frontend
 			if (right is NullExpression) {
 				StackFrame frame = context.CurrentFrame;
 				TargetType ltype = left.EvaluateType (context);
-				obj = frame.Language.CreateNullObject (frame.Thread, ltype);
+				obj = context.CurrentLanguage.CreateNullObject (frame.Thread, ltype);
 			} else
 				obj = right.EvaluateObject (context);
 
@@ -3340,7 +3238,7 @@ namespace Mono.Debugger.Frontend
 			if (obj == null)
 				return null;
 
-			TargetStructObject sobj = (TargetStructObject) obj;
+			TargetClassObject sobj = (TargetClassObject) obj;
 			for (int i = 0; i <= level; i++) {
 				sobj = sobj.GetParentObject (context.CurrentThread);
 				if (sobj == null)
@@ -3350,4 +3248,49 @@ namespace Mono.Debugger.Frontend
 			return sobj;
 		}
 	}
+
+	internal class TypeProxyExpression : Expression
+	{
+		string proxy_type;
+		TargetStructObject instance;
+
+		public TypeProxyExpression (string proxy_type, TargetStructObject instance)
+		{
+			this.proxy_type = proxy_type;
+			this.instance = instance;
+		}
+
+		public override string Name {
+			get { return String.Format ("[DebuggerTypeProxy(typeof({0}))]", proxy_type); }
+		}
+
+		TargetType LookupType (ScriptingContext context)
+		{
+			TargetType type = context.CurrentLanguage.LookupType (proxy_type);
+			if (type != null)
+				return type;
+
+			string[] namespaces = context.GetNamespaces () ?? new string [0];
+			foreach (string ns in namespaces) {
+				string full_name = SimpleNameExpression.MakeFQN (ns, proxy_type);
+				type = context.CurrentLanguage.LookupType (full_name);
+				if (type != null)
+					return type;
+			}
+
+			return null;
+		}
+
+		protected override Expression DoResolve (ScriptingContext context)
+		{
+			TargetType type = LookupType (context);
+			if (type == null)
+				return null;
+
+			Expression[] args = { new ArgumentExpression (instance) };
+			NewExpression expr = new NewExpression (new TypeExpression (type), args);
+
+			return expr.Resolve (context);
+		}
+	} 
 }
