@@ -13,12 +13,11 @@ using System.Runtime.Serialization.Formatters.Binary;
 
 using Mono.Debugger;
 using Mono.Debugger.Languages;
-
-using Mono.GetOptions;
+using EE=Mono.Debugger.ExpressionEvaluator;
 
 namespace Mono.Debugger.Frontend
 {
-	public class Interpreter : DebuggerMarshalByRefObject, IDisposable
+	public class Interpreter : DebuggerMarshalByRefObject, IInterruptionHandler, IDisposable
 	{
 		DebuggerConfiguration config;
 		DebuggerSession session;
@@ -61,6 +60,8 @@ namespace Mono.Debugger.Frontend
 			this.session = new DebuggerSession (config, options, "main", parser);
 			this.engine = new DebuggerEngine (this);
 
+			parser.Session = session;
+
 			source_factory = new SourceFileFactory ();
 
 			interrupt_event = new ManualResetEvent (false);
@@ -260,23 +261,10 @@ namespace Mono.Debugger.Frontend
 
 		public bool HasTarget {
 			get {
-				return debugger != null;
+				return (debugger != null) && debugger.HasTarget;
 			}
 		}
 
-		protected void CheckLastEvent (Thread thread)
-		{
-			if (thread == null)
-				return;
-
-			TargetEventArgs args = thread.GetLastTargetEvent ();
-			if (args == null)
-				return;
-
-			if ((args.Type == TargetEventType.TargetStopped) && ((int) args.Data == 0))
-				Style.TargetEvent (thread, args);
-		}
-
 		public Process Start ()
 		{
 			if ((debugger != null) || (main_process != null))
@@ -290,12 +278,12 @@ namespace Mono.Debugger.Frontend
 				debugger = new Debugger (config);
 
 				new InterpreterEventSink (this, debugger);
-				new ThreadEventSink (this, debugger);
-
-				current_process = main_process = debugger.Run (session);
 
+				CommandResult result;
+				current_process = main_process = debugger.Run (session, out result);
 				current_thread = current_process.MainThread;
-				current_thread = WaitAll (current_thread);
+
+				Wait (result);
 
 				return current_process;
 			} catch (TargetException) {
@@ -317,11 +305,12 @@ namespace Mono.Debugger.Frontend
 				debugger = new Debugger (config);
 
 				new InterpreterEventSink (this, debugger);
-				new ThreadEventSink (this, debugger);
 
-				current_process = main_process = debugger.Attach (session, pid);
+				CommandResult result;
+				current_process = main_process = debugger.Attach (session, pid, out result);
 				current_thread = current_process.MainThread;
-				current_thread = WaitAll (current_thread);
+
+				Wait (result);
 
 				return current_process;
 			} catch (TargetException) {
@@ -331,6 +320,7 @@ namespace Mono.Debugger.Frontend
 			}
 		}
 
+#if DISABLED
 		public Process OpenCoreFile (string core_file)
 		{
 			if ((debugger != null) || (main_process != null))
@@ -342,7 +332,6 @@ namespace Mono.Debugger.Frontend
 				debugger = new Debugger (config);
 
 				new InterpreterEventSink (this, debugger);
-				new ThreadEventSink (this, debugger);
 
 				Thread[] threads;
 				current_process = main_process = debugger.OpenCoreFile (
@@ -357,6 +346,7 @@ namespace Mono.Debugger.Frontend
 				throw;
 			}
 		}
+#endif
 
 		public void SaveSession (Stream stream)
 		{
@@ -370,15 +360,17 @@ namespace Mono.Debugger.Frontend
 
 			try {
 				debugger = new Debugger (config);
+				parser = new ExpressionParser (this);
 				session = new DebuggerSession (config, stream, parser);
+				parser.Session = session;
 
 				new InterpreterEventSink (this, debugger);
-				new ThreadEventSink (this, debugger);
-
-				current_process = main_process = debugger.Run (session);
 
+				CommandResult result;
+				current_process = main_process = debugger.Run (session, out result);
 				current_thread = current_process.MainThread;
-				current_thread = WaitAll (current_thread);
+
+				Wait (result);
 
 				return current_process;
 			} catch (TargetException ex) {
@@ -418,161 +410,6 @@ namespace Mono.Debugger.Frontend
 			return true;
 		}
 
-		public bool WaitOne (Thread thread)
-		{
-			ClearInterrupt ();
-
-			WaitHandle[] handles = new WaitHandle [2];
-			handles [0] = interrupt_event;
-			handles [1] = thread.WaitHandle;
-
-			int ret = WaitHandle.WaitAny (handles);
-
-			if (ret == 0) {
-				thread.Stop ();
-				thread.WaitHandle.WaitOne ();
-				return false;
-			}
-
-			CheckLastEvent (thread);
-			return true;
-		}
-
-		public Thread WaitAll (Thread thread)
-		{
-			return WaitAll (thread, thread.GetWaitHandle (), false);
-		}
-
-		public Thread WaitAll (Thread thread, CommandResult result, bool wait)
-		{
-			if (result == null)
-				return null;
-
-			ClearInterrupt ();
-			process_event.Reset ();
-
-			Thread stopped;
-			Hashtable seen_threads = new Hashtable ();
-			Hashtable wait_list = new Hashtable ();
-
-			do {
-			again:
-				Process[] processes = HasTarget ? debugger.Processes : new Process [0];
-				foreach (Process process in processes) {
-					Thread[] the_threads = process.GetThreads ();
-					foreach (Thread t in the_threads) {
-						if ((t == thread) || seen_threads.Contains (t))
-							continue;
-
-						seen_threads.Add (t, null);
-						if (t.IsRunning)
-							wait_list.Add (t, false);
-						else if ((t.ThreadFlags & Thread.Flags.AutoRun) != 0) {
-							t.Continue ();
-							wait_list.Add (t, true);
-						}
-					}
-				}
-
-				WaitHandle[] handles = new WaitHandle [wait_list.Count + 3];
-				handles [0] = interrupt_event;
-				handles [1] = process_event;
-				handles [2] = result.CompletedEvent;
-
-				Thread[] threads = new Thread [wait_list.Count];
-				wait_list.Keys.CopyTo (threads, 0);
-
-				for (int i = 0; i < wait_list.Count; i++)
-					handles [i + 3] = threads [i].WaitHandle;
-
-				int ret = WaitHandle.WaitAny (handles);
-
-				if (ret == 0) {
-					stopped = null;
-					result.Abort ();
-					result.CompletedEvent.WaitOne ();
-					break;
-				} else if (ret == 1) {
-					seen_threads = new Hashtable ();
-					wait_list = new Hashtable ();
-					process_event.Reset ();
-					goto again;
-				}
-
-				if (result.Result is Exception)
-					throw (Exception) result.Result;
-
-				if (ret == 2)
-					stopped = thread;
-				else
-					stopped = threads [ret - 3];
-
-				CheckLastEvent (stopped);
-				wait_list.Remove (stopped);
-			} while ((stopped != null) && (stopped != thread) && (wait || !stopped.IsAlive));
-
-			if (!HasTarget)
-				return stopped;
-
-			foreach (Process process in Processes) {
-				foreach (Thread t in process.GetThreads ()) {
-					if (t == stopped)
-						continue;
-					// Never touch immutable threads.
-					if (!t.IsRunning)
-						continue;
-					if ((t.ThreadFlags & Thread.Flags.Immutable) != 0)
-						continue;
-					// Background thread -> keep running.
-					if ((t.ThreadFlags & Thread.Flags.Background) != 0)
-						continue;
-
-					// Stop and set AutoRun.
-					t.AutoStop ();
-				}
-			}
-
-			return stopped;
-		}
-
-		protected void Wait (Thread thread)
-		{
-			if (thread == null)
-				return;
-
-			WaitHandle[] handles = new WaitHandle [2];
-			handles [0] = interrupt_event;
-			handles [1] = thread.WaitHandle;
-
-			WaitHandle.WaitAny (handles);
-		}
-
-		public RuntimeInvokeResult RuntimeInvoke (Thread thread,
-							  TargetFunctionType function,
-							  TargetStructObject object_argument,
-							  TargetObject[] param_objects,
-							  bool is_virtual, bool debug)
-		{
-			RuntimeInvokeResult result = thread.RuntimeInvoke (
-				function, object_argument, param_objects, is_virtual, debug);
-
-			if (DebuggerConfiguration.BrokenThreading) {
-				Thread ret = WaitAll (thread, result, false);
-				if (ret != thread) {
-					result.Abort ();
-					return null;
-				}
-			} else if (debug) {
-				Wait (thread);
-
-				CheckLastEvent (thread);
-			} else {
-				Wait (result);
-			}
-
-			return result;
-		}
-
 		public int Interrupt ()
 		{
 			interrupt_event.Set ();
@@ -585,6 +422,15 @@ namespace Mono.Debugger.Frontend
 			interrupt_event.Reset ();
 		}
 
+		WaitHandle IInterruptionHandler.InterruptionEvent {
+			get { return interrupt_event; }
+		}
+
+		bool IInterruptionHandler.CheckInterruption ()
+		{
+			return interrupt_level > 0;
+		}
+
 		public void ShowBreakpoints ()
 		{
 			Event[] events = Session.Events;
@@ -624,14 +470,14 @@ namespace Mono.Debugger.Frontend
 		protected virtual void OnThreadCreated (Thread thread)
 		{
 			process_event.Set ();
-			if (DebuggerConfiguration.NotifyUser_ThreadCreation)
+			if ((DebuggerConfiguration.UserNotifications & DebuggerConfiguration.UserNotificationType.Threads) != 0)
 				Print ("Process #{0} created new thread @{1}.",
 				       thread.Process.ID, thread.ID);
 		}
 
 		protected virtual void OnThreadExited (Thread thread)
 		{
-			if (DebuggerConfiguration.NotifyUser_ThreadCreation)
+			if ((DebuggerConfiguration.UserNotifications & DebuggerConfiguration.UserNotificationType.Threads) != 0)
 				Print ("Thread @{0} exited.", thread.ID);
 			if (thread == current_thread)
 				current_thread = null;
@@ -676,12 +522,11 @@ namespace Mono.Debugger.Frontend
 
 		protected virtual void OnTargetEvent (Thread thread, TargetEventArgs args)
 		{
-			if (args.Type == TargetEventType.TargetInterrupted) {
-				;
-			} else if ((args.Type != TargetEventType.TargetStopped) || ((int) args.Data != 0))
-				Style.TargetEvent (thread, args);
-			else if ((thread.ThreadFlags & Thread.Flags.Background) != 0)
-				Style.TargetEvent (thread, args);
+			if ((args.Type != TargetEventType.TargetSignaled) &&
+			    (args.Type != TargetEventType.TargetExited) &&
+			    (args.Type != TargetEventType.TargetInterrupted))
+				CurrentThread = thread;
+			Style.TargetEvent (thread, args);
 		}
 
 		protected virtual void OnTargetOutput (bool is_stderr, string line)
@@ -824,10 +669,10 @@ namespace Mono.Debugger.Frontend
 		}
 
 		public int InsertExceptionCatchPoint (Thread target, ThreadGroup group,
-						      TargetType exception)
+						      TargetType exception, bool unhandled)
 		{
 			Event handle = target.Process.Session.InsertExceptionCatchPoint (
-				target, group, exception);
+				target, group, exception, unhandled);
 			return handle.Index;
 		}
 
@@ -994,6 +839,22 @@ namespace Mono.Debugger.Frontend
 				context.ShowDisplay (d);
 		}
 
+		internal CommandLineInterpreter CLI {
+			get; set;
+		}
+
+		protected virtual void OnEnterNestedBreakState (Thread thread)
+		{
+			if (CLI != null)
+				CLI.EnterNestedBreakState ();
+		}
+
+		protected void OnLeaveNestedBreakState (Thread thread)
+		{
+			if (CLI != null)
+				CLI.LeaveNestedBreakState ();
+		}
+
 		//
 		// IDisposable
 		//
@@ -1041,20 +902,45 @@ namespace Mono.Debugger.Frontend
 
 				debugger.TargetExitedEvent += target_exited;
 				debugger.ThreadCreatedEvent += thread_created;
+				debugger.ManagedThreadCreatedEvent += managed_thread_created;
 				debugger.ThreadExitedEvent += thread_exited;
 				debugger.MainProcessCreatedEvent += main_process_created;
 				debugger.ProcessReachedMainEvent += process_reached_main;
 				debugger.ProcessCreatedEvent += process_created;
 				debugger.ProcessExitedEvent += process_exited;
 				debugger.ProcessExecdEvent += process_execd;
+				debugger.TargetEvent += target_event;
+				debugger.EnterNestedBreakStateEvent +=
+					delegate (Debugger unused, Thread thread) {
+						interpreter.OnEnterNestedBreakState (thread);
+					};
+				debugger.LeaveNestedBreakStateEvent +=
+					delegate (Debugger unused, Thread thread) {
+						interpreter.OnLeaveNestedBreakState (thread);
+					};
+			}
+
+			public void target_event (Thread thread, TargetEventArgs args)
+			{
+				if (((thread.ThreadFlags & Thread.Flags.Daemon) != 0) &&
+				    ((args.Type == TargetEventType.TargetExited) ||
+				     (args.Type == TargetEventType.TargetSignaled)))
+					return;
+				interpreter.OnTargetEvent (thread, args);
 			}
 
 			public void thread_created (Debugger debugger, Thread thread)
 			{
-				if ((thread.ThreadFlags & Thread.Flags.Daemon) == 0)
+				if (!thread.Process.IsManaged ||
+				    ((thread.ThreadFlags & Thread.Flags.Daemon) == 0))
 					interpreter.OnThreadCreated (thread);
 			}
 
+			public void managed_thread_created (Debugger debugger, Thread thread)
+			{
+				interpreter.OnThreadCreated (thread);
+			}
+
 			public void thread_exited (Debugger debugger, Thread thread)
 			{
 				if ((thread.ThreadFlags & Thread.Flags.Daemon) == 0)
@@ -1107,27 +993,5 @@ namespace Mono.Debugger.Frontend
 				interpreter.OnTargetOutput (is_stderr, line);
 			}
 		}
-
-		[Serializable]
-		protected class ThreadEventSink
-		{
-			Interpreter interpreter;
-
-			public ThreadEventSink (Interpreter interpreter, Debugger debugger)
-			{
-				this.interpreter = interpreter;
-
-				debugger.TargetEvent += target_event;
-			}
-
-			public void target_event (Thread thread, TargetEventArgs args)
-			{
-				if (((thread.ThreadFlags & Thread.Flags.Daemon) != 0) &&
-				    ((args.Type == TargetEventType.TargetExited) ||
-				     (args.Type == TargetEventType.TargetSignaled)))
-					return;
-				interpreter.OnTargetEvent (thread, args);
-			}
-		}
 	}
 }
