@@ -1,7 +1,9 @@
 using System;
 using System.Collections;
 using System.Collections.Generic;
+
 using Mono.Debugger.Backend;
+using Mono.Debugger.Backend.Mono;
 
 namespace Mono.Debugger.Languages.Mono
 {
@@ -30,29 +32,42 @@ namespace Mono.Debugger.Languages.Mono
 							   TargetMemoryAccess target,
 							   TargetAddress klass)
 		{
-			TargetAddress image = mono.MonoRuntime.MonoClassGetMonoImage (target, klass);
+			TargetAddress image = mono.MetadataHelper.MonoClassGetMonoImage (target, klass);
 			MonoSymbolFile file = mono.GetImage (image);
 			if (file == null)
 				throw new InternalError ();
 
-			int token = mono.MonoRuntime.MonoClassGetToken (target, klass);
+			int token = mono.MetadataHelper.MonoClassGetToken (target, klass);
 			if ((token & 0xff000000) != 0x02000000)
 				throw new InternalError ();
 
 			Cecil.TypeDefinition typedef;
-			typedef = (Cecil.TypeDefinition) file.ModuleDefinition.LookupByToken (
-				Cecil.Metadata.TokenType.TypeDef, token & 0x00ffffff);
+			typedef = (Cecil.TypeDefinition) file.ModuleDefinition.LookupToken (
+				new Cecil.MetadataToken (Cecil.TokenType.TypeDef, token & 0x00ffffff));
 			if (typedef == null)
 				throw new InternalError ();
 
 			MonoClassInfo info = new MonoClassInfo (file, typedef, target, klass);
-			if (info.IsGenericClass) {
-				info.struct_type = file.MonoLanguage.ReadGenericClass (
-					target, info.GenericClass);
+			if ((file == mono.BuiltinTypes.Corlib) &&
+			    (typedef.FullName == "System.Decimal")) {
+				MonoFundamentalType ftype = mono.BuiltinTypes.DecimalType;
+
+				if (ftype.ClassType == null) {
+					MonoClassType ctype = new MonoClassType (file, typedef, info);
+					((IMonoStructType) ctype).ClassInfo = info;
+					ftype.SetClass (ctype);
+				}
+
+				info.struct_type = (IMonoStructType) ftype.ClassType;
+				info.type = ftype;
+			} else if (info.IsGenericClass) {
+				info.struct_type = (IMonoStructType)
+					file.MonoLanguage.ReadGenericClass (
+						target, info.GenericClass, false);
 				info.type = info.struct_type.Type;
 			} else {
 				info.type = file.LookupMonoType (typedef);
-				if (info.type is TargetStructType)
+				if (info.type is TargetClassType)
 					info.struct_type = (IMonoStructType) info.type;
 				else
 					info.struct_type = (IMonoStructType) info.type.ClassType;
@@ -82,20 +97,20 @@ namespace Mono.Debugger.Languages.Mono
 			this.KlassAddress = klass;
 			this.CecilType = typedef;
 
-			parent_klass = MonoRuntime.MonoClassGetParent (target, klass);
-			GenericClass = MonoRuntime.MonoClassGetGenericClass (target, klass);
-			GenericContainer = MonoRuntime.MonoClassGetGenericContainer (target, klass);
+			parent_klass = MetadataHelper.MonoClassGetParent (target, klass);
+			GenericClass = MetadataHelper.MonoClassGetGenericClass (target, klass);
+			GenericContainer = MetadataHelper.MonoClassGetGenericContainer (target, klass);
 		}
 
-		protected MonoRuntime MonoRuntime {
-			get { return SymbolFile.MonoLanguage.MonoRuntime; }
+		protected MetadataHelper MetadataHelper {
+			get { return SymbolFile.MonoLanguage.MetadataHelper; }
 		}
 
 		public override TargetType RealType {
 			get { return type; }
 		}
 
-		public override TargetStructType Type {
+		public override TargetClassType Type {
 			get { return struct_type.Type; }
 		}
 
@@ -103,12 +118,19 @@ namespace Mono.Debugger.Languages.Mono
 			get { return !GenericClass.IsNull; }
 		}
 
+		internal int GetInstanceSize (TargetMemoryAccess target)
+		{
+			return MetadataHelper.MonoClassGetInstanceSize (target, KlassAddress);
+		}
+
 		internal MonoFieldInfo[] GetFields (TargetMemoryAccess target)
 		{
 			if (fields != null)
 				return fields;
 
-			int field_count = MonoRuntime.MonoClassGetFieldCount (target, KlassAddress);
+			int field_count = MetadataHelper.MonoClassGetFieldCount (target, KlassAddress);
+			if ((field_count != 0) && !MetadataHelper.MonoClassHasFields (target, KlassAddress))
+				throw new TargetException (TargetError.ClassNotInitialized);
 
 			fields = new MonoFieldInfo [field_count];
 			field_offsets = new int [field_count];
@@ -117,11 +139,11 @@ namespace Mono.Debugger.Languages.Mono
 			for (int i = 0; i < field_count; i++) {
 				Cecil.FieldDefinition field = CecilType.Fields [i];
 
-				TargetAddress type_addr = MonoRuntime.MonoClassGetFieldType (
+				TargetAddress type_addr = MetadataHelper.MonoClassGetFieldType (
 					target, KlassAddress, i);
 
 				field_types [i] = SymbolFile.MonoLanguage.ReadType (target, type_addr);
-				field_offsets [i] = MonoRuntime.MonoClassGetFieldOffset (
+				field_offsets [i] = MetadataHelper.MonoClassGetFieldOffset (
 					target, KlassAddress, i);
 
 				fields [i] = new MonoFieldInfo (struct_type, field_types [i], i, field);
@@ -185,13 +207,16 @@ namespace Mono.Debugger.Languages.Mono
 			TargetAddress addr = field_loc.GetAddress (target);
 
 			if (field_loc.HasAddress && field_loc.GetAddress (target).IsNull)
-				return new MonoNullObject (type, field_loc);
+				return new TargetNullObject (type);
 
 			return type.GetObject (target, field_loc);
 		}
 
 		internal TargetObject GetStaticField (Thread thread, TargetFieldInfo field)
 		{
+			if (!thread.CurrentFrame.Language.IsManaged)
+				throw new TargetException (TargetError.InvalidContext);
+
 			TargetAddress data_address = thread.CallMethod (
 				SymbolFile.MonoLanguage.MonoDebuggerInfo.ClassGetStaticFieldData,
 				KlassAddress, 0);
@@ -261,6 +286,9 @@ namespace Mono.Debugger.Languages.Mono
 		internal void SetStaticField (Thread thread, TargetFieldInfo field,
 					      TargetObject obj)
 		{
+			if (!thread.CurrentFrame.Language.IsManaged)
+				throw new TargetException (TargetError.InvalidContext);
+
 			TargetAddress data_address = thread.CallMethod (
 				SymbolFile.MonoLanguage.MonoDebuggerInfo.ClassGetStaticFieldData,
 				KlassAddress, 0);
@@ -318,17 +346,17 @@ namespace Mono.Debugger.Languages.Mono
 				return;
 
 			try {
-				if (!MonoRuntime.MonoClassHasMethods (target, KlassAddress))
+				if (!MetadataHelper.MonoClassHasMethods (target, KlassAddress))
 					return;
 
-				int count = MonoRuntime.MonoClassGetMethodCount (target, KlassAddress);
+				int count = MetadataHelper.MonoClassGetMethodCount (target, KlassAddress);
 
 				methods_by_token = new Dictionary<int,TargetAddress> ();
 
 				for (int i = 0; i < count; i++) {
-					TargetAddress address = MonoRuntime.MonoClassGetMethod (
+					TargetAddress address = MetadataHelper.MonoClassGetMethod (
 						target, KlassAddress, i);
-					int mtoken = MonoRuntime.MonoMethodGetToken (target, address);
+					int mtoken = MetadataHelper.MonoMethodGetToken (target, address);
 					if (mtoken != 0)
 						methods_by_token.Add (mtoken, address);
 				}
@@ -369,7 +397,7 @@ namespace Mono.Debugger.Languages.Mono
 
 		void get_parent (TargetMemoryAccess target)
 		{
-			parent_klass = MonoRuntime.MonoClassGetParent (target, KlassAddress);
+			parent_klass = MetadataHelper.MonoClassGetParent (target, KlassAddress);
 			if (parent_klass.IsNull)
 				return;
 
