@@ -1,7 +1,12 @@
 using System;
 using System.Text;
+using System.Diagnostics;
+using System.Collections;
 using Cecil = Mono.Cecil;
 
+using Mono.Debugger.Backend;
+using Mono.Debugger.Backend.Mono;
+
 namespace Mono.Debugger.Languages.Mono
 {
 	[Serializable]
@@ -9,10 +14,10 @@ namespace Mono.Debugger.Languages.Mono
 	{
 		[NonSerialized]
 		public readonly Cecil.FieldDefinition FieldInfo;
+		DebuggerBrowsableState? browsable_state = null;
+		DebuggerDisplayAttribute debugger_display;
 		bool is_compiler_generated;
 
-		const string cgen_attr = "System.Runtime.CompilerServices.CompilerGeneratedAttribute";
-
 		public MonoFieldInfo (IMonoStructType type, TargetType field_type, int pos,
 				      Cecil.FieldDefinition finfo)
 			: base (field_type, finfo.Name, pos, finfo.IsStatic,
@@ -20,11 +25,20 @@ namespace Mono.Debugger.Languages.Mono
 		{
 			FieldInfo = finfo;
 
-			foreach (Cecil.CustomAttribute cattr in finfo.CustomAttributes) {
-				string cname = cattr.Constructor.DeclaringType.FullName;
-				if (cname == cgen_attr)
-					is_compiler_generated = true;
-			}
+			DebuggerTypeProxyAttribute type_proxy;
+			MonoSymbolFile.CheckCustomAttributes (finfo,
+							      out browsable_state,
+							      out debugger_display,
+							      out type_proxy,
+							      out is_compiler_generated);
+		}
+
+		public override DebuggerBrowsableState? DebuggerBrowsableState {
+			get { return browsable_state; }
+		}
+
+		public override DebuggerDisplayAttribute DebuggerDisplayAttribute {
+			get { return debugger_display; }
 		}
 
 		internal static TargetMemberAccessibility GetAccessibility (Cecil.FieldDefinition field)
@@ -110,7 +124,7 @@ namespace Mono.Debugger.Languages.Mono
 			case Cecil.MethodAttributes.Family:
 			case Cecil.MethodAttributes.FamANDAssem:
 				return TargetMemberAccessibility.Protected;
-			case Cecil.MethodAttributes.Assem:
+			case Cecil.MethodAttributes.Assembly:
 			case Cecil.MethodAttributes.FamORAssem:
 				return TargetMemberAccessibility.Internal;
 			default:
@@ -122,10 +136,10 @@ namespace Mono.Debugger.Languages.Mono
 	[Serializable]
 	internal class MonoEventInfo : TargetEventInfo
 	{
-		public readonly MonoClassType Klass;
+		public readonly IMonoStructType Klass;
 		public readonly MonoFunctionType AddType, RemoveType, RaiseType;
 
-		private MonoEventInfo (MonoClassType klass, int index, Cecil.EventDefinition einfo,
+		private MonoEventInfo (IMonoStructType klass, int index, Cecil.EventDefinition einfo,
 				       TargetType type, bool is_static,
 				       TargetMemberAccessibility accessibility, MonoFunctionType add,
 				       MonoFunctionType remove, MonoFunctionType raise)
@@ -137,7 +151,7 @@ namespace Mono.Debugger.Languages.Mono
 			this.RaiseType = raise;
 		}
 
-		internal static MonoEventInfo Create (MonoClassType klass, int index,
+		internal static MonoEventInfo Create (IMonoStructType klass, int index,
 						      Cecil.EventDefinition einfo)
 		{
 			TargetType type = klass.File.MonoLanguage.LookupMonoType (einfo.EventType);
@@ -177,6 +191,8 @@ namespace Mono.Debugger.Languages.Mono
 	{
 		public readonly IMonoStructType Klass;
 		public readonly MonoFunctionType GetterType, SetterType;
+		DebuggerBrowsableState? browsable_state = null;
+		DebuggerDisplayAttribute debugger_display;
 
 		private MonoPropertyInfo (TargetType type, IMonoStructType klass, int index,
 					  bool is_static, Cecil.PropertyDefinition pinfo,
@@ -187,6 +203,22 @@ namespace Mono.Debugger.Languages.Mono
 			this.Klass = klass;
 			this.GetterType = getter;
 			this.SetterType = setter;
+
+			bool is_compiler_generated;
+			DebuggerTypeProxyAttribute type_proxy;
+			MonoSymbolFile.CheckCustomAttributes (pinfo,
+							      out browsable_state,
+							      out debugger_display,
+							      out type_proxy,
+							      out is_compiler_generated);
+		}
+
+		public override DebuggerBrowsableState? DebuggerBrowsableState {
+			get { return browsable_state; }
+		}
+
+		public override DebuggerDisplayAttribute DebuggerDisplayAttribute {
+			get { return debugger_display; }
 		}
 
 		internal static MonoPropertyInfo Create (IMonoStructType klass, int index,
