@@ -4,11 +4,12 @@ using NUnit.Framework;
 using Mono.Debugger;
 using Mono.Debugger.Languages;
 using Mono.Debugger.Frontend;
+using Mono.Debugger.Test.Framework;
 
 namespace Mono.Debugger.Tests
 {
-	[TestFixture]
-	public class TestInvocation : TestSuite
+	[DebuggerTestFixture]
+	public class TestInvocation : DebuggerTestFixture
 	{
 		public TestInvocation ()
 			: base ("TestInvocation")
@@ -79,7 +80,7 @@ namespace Mono.Debugger.Tests
 					     "X.BreakpointTest()", GetLine ("breakpoint test"));
 
 			AssertExecute ("continue");
-			AssertStopped (thread, "main2", "X.Main()");
+			AssertRuntimeInvokeDone (thread, "X.Main()", GetLine ("main2"));
 			AssertExecute ("continue");
 
 			AssertHitBreakpoint (thread, bpt_breakpoint_test,
